
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
import numpy as np
import pandas as pd
from dataclasses import dataclass, field
import colorcet as cc
from plottable import Table, ColDef

import matplotlib.cm as mplc
import matplotlib.dates as mdates
import matplotlib.transforms as mtr
import matplotlib.artist as mpla
import matplotlib.axes as mplax
from matplotlib.ticker import MaxNLocator
from matplotlib.lines import Line2D
import matplotlib.patches as mpatches
from cartopy.mpl.geoaxes import GeoAxes
from cartopy.mpl.ticker import LongitudeLocator, LatitudeLocator
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from matplotlib.colors import LogNorm, PowerNorm, Colormap, ListedColormap
from typing import Iterable, Literal, Sequence

@dataclass(kw_only = True)
class base:

    ax: plt.Axes

    grid: bool = True 
    grid_which: Literal['major', 'minor', 'both'] = 'major'
    grid_color: str = 'dimgray'
    grid_ls: str = '--'
    grid_alpha: float = 0.8
    tick_pad: float = 10.0

    no_spines: None | list[str] = field(default_factory = 
                                        lambda: ['top', 
                                                 'right', 
                                                 'bottom', 
                                                 'left'])

    y_title: float = 1.0
    fs_title: int = 10 
    title: str = '' 

    axis: bool = False
    axis_color: str = 'k' 
    axis_ls: str = '--' 
    axis_lw: float = 0.7 
    axis_alpha: float = 0.8 
    axis_dashes: tuple = (4, 4)

    xlabel: str = '' 
    ylabel: str = '' 

    fs_label: float = 12.0
    ax_tag: str = ''
    ax_tag_x: float = 0.5
    ax_tag_y: float = 1.0 

    ax_tag_ha: Literal['left', 'center', 'right'] = 'center'
    ax_tag_va: Literal['bottom', 'center', 'top'] = 'center'


    def decoration(self):

        self.plot_projection = self.ax.transData
        
        if self.grid:
        
            self.ax.grid(which = self.grid_which, 
                         color = self.grid_color, 
                         visible = True, 
                         ls = self.grid_ls, 
                         alpha = self.grid_alpha, 
                         zorder = 0)


        if self.no_spines is not None:
            for spine in self.no_spines:
                    self.ax.spines[spine].set_visible(False)
            
        self.ax.set_title(self.title, 
                          fontsize = self.fs_title, 
                          y = self.y_title)
        
        self.ax.set_ylabel(self.ylabel, 
                           fontsize = self.fs_label)
        
        self.ax.set_xlabel(self.xlabel, 
                           fontsize = self.fs_label)

        self.ax.tick_params(axis = 'both', 
                            which = 'major', 
                            pad = self.tick_pad)
        
        if self.axis:

            self.ax.axline((0, 0), (0, 1),
                           color = self.axis_color, 
                           ls = self.axis_ls, 
                           lw = self.axis_lw, 
                           alpha = self.axis_alpha, 
                           dashes = self.axis_dashes,
                           transform = self.ax.transAxes, 
                           zorder = 0)

            self.ax.axline((0, 0), (1, 0),
                           color = self.axis_color, 
                           ls = self.axis_ls, 
                           lw = self.axis_lw, 
                           alpha = self.axis_alpha, 
                           dashes = self.axis_dashes, 
                           transform = self.ax.transAxes,
                           zorder = 0)
        
        props = dict(facecolor = 'white', 
                     edgecolor = 'none', 
                     alpha = 0.6)

        self.ax.text(self.ax_tag_x, 
                     self.ax_tag_y, 
                     self.ax_tag, 
                     fontsize = self.fs_label, 
                     transform = self.ax.transAxes, 
                     va = self.ax_tag_va, 
                     ha = self.ax_tag_ha,
                     bbox = props,
                     zorder = 8)

@dataclass(kw_only = True)
class base_001(base):

    def colormesh(self,
                  lon: np.ndarray,
                  lat: np.ndarray,
                  array: np.ndarray,
                  cmap: str | Colormap | ListedColormap = 'viridis',
                  cmap_n: int = 1000,
                  vmin: float | None = None,
                  vmax: float | None = None,
                  norm: Literal['log'] | None = None,
                  alpha: float = 1.0,
                  zorder: int = 0):
        
        if norm == 'log':

            norm_a = LogNorm(vmin = vmin,
                             vmax = vmax)
            
            vmin_a = None
            vmax_a = None

        elif norm == 'power':

            norm_a = PowerNorm(vmin = vmin,
                               vmax = vmax)
            
            vmin_a = None
            vmax_a = None

        else:

            norm_a = norm
            vmin_a = vmin
            vmax_a = vmax

        if isinstance(cmap, str):
            
            cmapn = mplc.get_cmap(cmap, cmap_n)

        elif isinstance(cmap, ListedColormap):

            cmapn = cmap
        
        artist = self.ax.pcolormesh(lon, 
                                    lat, 
                                    array, 
                                    cmap = cmapn, 
                                    vmin = vmin_a, 
                                    vmax = vmax_a, 
                                    transform = self.plot_projection, 
                                    zorder = zorder,
                                    norm = norm_a,
                                    alpha = alpha)
    
        return artist
    

    def contourf(self,
                 lon: np.ndarray,
                 lat: np.ndarray,
                 array: np.ndarray,
                 levels: list[float],
                 hatches: list[str],
                 colors: str | list[str],
                 extend: Literal['neither', 'both', 'min', 'max'] = 'neither',
                 cmap: None | str | Colormap | ListedColormap = None,
                 cmap_n: int = 1000,
                 vmin: float | None = None,
                 vmax: float | None = None,
                 alpha: None | float = 0.8,
                 zorder: int = 5):
        

        if isinstance(cmap, str):
            
            cmapn = mplc.get_cmap(cmap, cmap_n)

        elif isinstance(cmap, ListedColormap) or (cmap is None):

            cmapn = cmap
    
        
        artist = self.ax.contourf(lon,
                                  lat,
                                  array,
                                  cmap = cmapn,
                                  levels = levels,
                                  hatches = hatches,
                                  colors = colors,
                                  extend = extend,
                                  vmin = vmin,
                                  vmax = vmax,
                                  transform = self.plot_projection,
                                  alpha = alpha,
                                  zorder = zorder)
        
        return artist
    

    def scatter(self,
                lats: np.ndarray | list[float],
                lons: np.ndarray | list[float],
                color: str | np.ndarray | list[float] | list[str] = 'k',
                cmap: str  = 'viridis',
                vmin: float | None = None,
                vmax: float | None = None,
                alpha: float = 0.8,
                markersize: float | np.ndarray = 3.0,
                marker: str = 'o',
                zorder: int = 5):

        if isinstance(color, str): color = [color] * len(lats)

        artist = self.ax.scatter(lons, 
                                 lats,
                                 s = markersize,
                                 c = color,
                                 cmap = cmap,
                                 vmin = vmin,
                                 vmax = vmax,
                                 transform = self.plot_projection, 
                                 marker = marker,
                                 alpha = alpha, 
                                 zorder = zorder)
        
        return artist
    
    def color_legend(self,
                     fig: Figure,
                     dict_lines: dict,
                     fs_labels: float = 10.0,
                     anchor: tuple = (0.5, 0), 
                     markerfirst: bool = True,
                     handletextpad: float = 0.35,
                     columnspacing: float = 0.9, 
                     loc: str = 'lower center', 
                     handlelength: float = 1.0,
                     ncols: int = 1):

        handles = [Line2D([0], [0], 
                          color = v['color'],
                          linewidth = v['lw'], 
                          linestyle = v['ls'],
                          label = k) 
                   for k, v in dict_lines.items()]

        labels = list(dict_lines.keys())

        fig.legend(handles, labels, 
                   fontsize = fs_labels, 
                   ncol = ncols, 
                   frameon = False, 
                   loc = loc, 
                   bbox_to_anchor = anchor, 
                   handletextpad = handletextpad, 
                   columnspacing = columnspacing,
                   handlelength = handlelength,  
                   markerfirst = markerfirst,
                   bbox_transform = fig.transFigure)
        
    
    def hatch_legend(self,
                     fig: Figure,
                     dict_hatch: dict,
                     fs_labels: float = 10.0,
                     anchor: tuple = (0.5, 0), 
                     markerfirst: bool = True,
                     handletextpad: float = 0.35,
                     columnspacing: float = 0.9, 
                     loc: str = 'lower center', 
                     handlelength: float = 1.0,
                     ncols: int = 1):

        handles = [mpatches.Patch(facecolor = v['facecolor'],
                                  hatch = v['hatch'],
                                  label = k) 
                   for k, v in dict_hatch.items()]

        labels = list(dict_hatch.keys())

        fig.legend(handles, labels, 
                   fontsize = fs_labels, 
                   ncol = ncols, 
                   frameon = False, 
                   loc = loc, 
                   bbox_to_anchor = anchor, 
                   handletextpad = handletextpad, 
                   columnspacing = columnspacing,
                   handlelength = handlelength,  
                   markerfirst = markerfirst,
                   bbox_transform = fig.transFigure)
        

    
    def colorbar(self,
                 artist: mpla.Artist,
                 ax: mplax.Axes | Iterable[mplax.Axes],
                 label: str = '',
                 pad: float = 0.05,
                 label_pad: float = 15.0,
                 shrink: float = 1.0,
                 fraction: float = 0.5,
                 fs_label: float = 9.0,
                 aspect: float = 10.0,
                 tick_labels: list | None = None,
                 orientation: Literal['vertical', 'horizontal'] = 'vertical',
                 location: Literal['left', 'right', 'top', 'bottom'] = 'right',
                 extend: Literal['both', 'neither', 'min', 'max'] = 'both',
                 label_rotation: float = 270.0):
        
        cbar = plt.colorbar(artist, 
                            ax = ax, 
                            extend = extend, 
                            shrink = shrink, 
                            fraction = fraction,
                            aspect = aspect,
                            orientation = orientation,
                            location = location,
                            pad = pad)
        
        cbar.ax.tick_params(labelsize = fs_label)

        cbar.outline.set_visible(False)

        len_tick_labels = len(tick_labels) if tick_labels is not None else 0

        if orientation == 'vertical':

            cbar.ax.set_ylabel(label, 
                               labelpad = label_pad, 
                               rotation = label_rotation, 
                               fontsize = fs_label)
            
            cbar.ax.yaxis.set_ticks_position('right')

        elif orientation == 'horizontal': 

            cbar.ax.set_xlabel(label, 
                               labelpad = label_pad, 
                               rotation = 0, 
                               fontsize = fs_label)
            
            cbar.ax.xaxis.set_ticks_position('bottom')
        
        if len_tick_labels > 0:

            positions = np.linspace(0.5, 
                                    (len_tick_labels - 1) - 0.5, 
                                    len_tick_labels)

            if orientation == 'vertical':
        
                cbar.ax.set_yticks(positions, 
                                   tick_labels, 
                                   fontsize = fs_label)

            if orientation == 'horizontal':
                cbar.ax.set_xticks(positions, 
                                   tick_labels, 
                                   fontsize = fs_label)
            
        return cbar
    

    def plot(self,
             xs: np.ndarray | pd.Series,
             ys: np.ndarray | pd.Series | Sequence[np.ndarray | pd.Series],
             colors: str | list[str] = 'k',
             style: str | list[str] = 'none',
             lw: float | list[float] = 1.0,
             alpha: float = 0.8,
             projection: None | mtr.Transform | ccrs.Projection = None,
             markersize: float = 3.0,
             marker: str | list[str] = 'o',
             fillstyle: str = 'full',
             zorder: int = 5):

        #ys = np.stack([y.flatten() for y in ys], axis = 1)
        
        if projection is None: 
            projection = self.plot_projection

        if isinstance(colors, str):
            colors = [colors]
        
        if isinstance(style, str):
            style = [style]
        
        if isinstance(marker, str):
            marker = [marker] 
            
        if isinstance(lw, int) or isinstance(lw, float):
            lw = [lw]

        self.ax.set_prop_cycle(color = colors,
                               linestyle = style,
                               marker = marker,
                               linewidth = lw)

        self.ax.plot(xs, 
                     ys,
                     transform = projection, 
                     markersize = markersize,
                     fillstyle = fillstyle,
                     alpha = alpha, 
                     zorder = zorder)
        

    def error(self,
             xs: np.ndarray | pd.Series,
             ys: np.ndarray | pd.Series,
             x_err: list[np.ndarray] | list[pd.Series],
             y_err: list[np.ndarray] | list[pd.Series],
             colors: str | list[str] = 'k',
             lw: float | list[float] = 1.0,
             alpha: float = 0.8,
             projection: None | mtr.Transform | ccrs.Projection = None,
             zorder: int = 5):

        ys = np.stack([y.flatten() for y in ys], axis = 1)
        
        if projection is None: 
            projection = self.plot_projection

        if isinstance(colors, str):
            colors = [colors]
            
        if isinstance(lw, int) or isinstance(lw, float):
            lw = [lw]

        self.ax.set_prop_cycle(color = colors,
                               linewidth = lw)

        self.ax.errorbar(xs, 
                         ys,
                         xerr = x_err,
                         yerr = y_err,
                         transform = projection,
                         alpha = alpha, 
                         zorder = zorder)
        
        
    def fillb(self,
              xs: np.ndarray | pd.Series,
              lower: np.ndarray | pd.Series,
              upper: np.ndarray | pd.Series,
              color: str | list[str] = 'k',
              alpha: float = 0.8,
              projection: None | mtr.Transform | ccrs.Projection = None,
              zorder: int = 3):
        
        if projection is None: 
            projection = self.ax.transData
            
        self.ax.fill_between(xs, lower, upper, 
                             color = color, 
                             transform = projection,
                             alpha = alpha,
                             zorder = zorder)
        
    def table(self,
              data: pd.DataFrame,
              index_col: str | None = None,
              fontsize: float = 6,
              ha: Literal['left', 'center', 'right'] = 'left',
              column_definitions: list[ColDef] | None = None):
        
        textprops = {'fontsize': fontsize,
                     'ha': ha}
        
        Table(data,
              ax = self.ax,
              textprops = textprops,
              index_col = index_col,
              column_definitions = column_definitions)
        
        self.ax.set_title(self.title)


    def boxplot(self,
                data: Sequence[np.ndarray | pd.Series],
                tick_labels: list[str], 
                patches: bool = True,
                fliers: bool = False,
                notch: bool = True,
                colors: str | list[str] = 'k',
                widths: float | list[float] = 0.56,
                ls: str = '-',
                lw: float | list[float] = 1.0,
                median_lw: float = 1.0,
                whisker_lw: float = 1.0,
                line_color = 'k',
                median_color = 'k', #'#FB9902',
                alpha: float = 0.8,
                zorder: int = 5):
        
        if isinstance(colors, str): colors = [colors] * len(data)
        
        boxprops = {'linestyle': ls,
                    'linewidth': lw,
                    'color': line_color,
                    'alpha': alpha,}
        
        medianprops = {'linewidth': median_lw,
                       'color': median_color}
        
        whiskerprops = {'linewidth': whisker_lw,
                       'color': line_color}

        boxp = self.ax.boxplot(data,
                               patch_artist = patches,
                               showfliers = fliers,
                               notch = notch,
                               widths = widths,
                               boxprops = boxprops,
                               medianprops = medianprops,
                               whiskerprops = whiskerprops,
                               labels = tick_labels,
                               zorder = zorder,)

        #for element in ['whiskers', 'fliers', 'means', 'medians', 'caps']:
        #    plt.setp(boxp[element], 
        #             color = line_color,
        #             alpha = alpha)
            
        for patch, color in zip(boxp['boxes'], colors):
            plt.setp(patch, 
                     facecolor = color,
                     alpha = alpha)

    

@dataclass(kw_only = True)
class time_series(base_001):
    
    xs: np.ndarray | pd.Series
    ys: np.ndarray | pd.Series | Sequence[np.ndarray | pd.Series]

    fs_ticks: float = 10.0
    ticks_y: bool = True
    ticks_x: bool = True
    nticks_y: int = 5
    nticks_x: int = 5
    integery: bool = False

    date_limits: None | list[np.datetime64] = None
    y_limits: None | list[float] = None
    y_ticks_rel_b: float = 0.1
    x_ticks_rel_b: float = 0.1

    def ticks(self):

        xlocator = mdates.AutoDateLocator(maxticks = self.nticks_x)
        ylocator = MaxNLocator(prune = 'both', 
                               nbins = self.nticks_y, 
                               integer = self.integery)
    
        self.ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(xlocator))

        self.ax.yaxis.set_major_locator(ylocator)
        self.ax.xaxis.set_major_locator(xlocator)

        if not self.ticks_x: self.ax.set_xticklabels([])
        if not self.ticks_y: self.ax.set_yticklabels([])

        self.ax.tick_params(axis = 'both', 
                            which = 'major', 
                            labelsize = self.fs_ticks)


    def limits(self):

        if self.date_limits is None:
        
            tsmax = np.max(self.xs)
            tsmin = np.min(self.xs)


            rel_d = (tsmax - tsmin).days * self.x_ticks_rel_b

            rel_dd = pd.Timedelta(rel_d, 'D')

            self.ax.set_xlim(tsmin - rel_dd, tsmax + rel_dd)
        
        else:

            for id, date in enumerate(self.date_limits):

                if isinstance(date, str):

                    self.date_limits[id] = pd.to_datetime(date).to_datetime64()

            self.ax.set_xlim(self.date_limits[0], 
                             self.date_limits[1])

            
        if self.y_limits is None:
            
            tsmax = np.nanmax(self.ys) 
            tsmin = np.nanmin(self.ys)

            tsabs = np.max(np.abs([tsmax, tsmin]))

            tsmax_rel = tsmax + tsabs * self.y_ticks_rel_b
            tsmin_rel = tsmin - tsabs * self.y_ticks_rel_b

            self.ax.set_ylim(tsmin_rel, tsmax_rel)

        else:

            self.ax.set_ylim(self.y_limits[0], 
                             self.y_limits[1])
            
    
    def create(self):

        super().decoration()
        
        self.ticks()
        #self.limits()

        return self


@dataclass(kw_only = True)
class amap(base_001):
    
    ax: GeoAxes 

    grid: bool = False

    lon_extents: list[float] = field(default_factory =
                                    lambda: [-180.0, 180.0])
    lat_extents: list[float] = field(default_factory =
                                     lambda: [-90, 90])

    semmj_axis: None | int = 6370000
    semmn_axis: None | int = 6370000 

    fs_ticks: float = 7.0
    ticks_y: bool = True
    ticks_x: bool = True
    nticks_y: int = 2
    nticks_x: int = 2
    integery: bool = True
    integerx: bool = True
    
    feature_land: bool = False
    feature_ocean: bool = False
    
    land_alpha: float = 0.7
    ocean_alpha: float = 0.7
    land_color: str = 'gray'
    ocean_color: str = 'steelblue'
    lw_coast: float = 0.8
    lw_lines: float = 0.8
    color_lines: str = 'grey'
    ls_lines: str = '--'
    label_lines: list[str] = field(default_factory = 
                                        lambda: ['right', 
                                                 'bottom'])
    
    def updates(self):
        
        self.plot_projection: ccrs.Projection = ccrs.PlateCarree()

        if isinstance(self.ax, plt.Axes): NotImplementedError('Wrong axis type')

    def features(self):

        if self.feature_ocean:
            self.ax.add_feature(cfeature.OCEAN,
                                color = self.ocean_color, 
                                alpha = self.ocean_alpha, 
                                zorder = 0)
            
        if self.feature_land:
            self.ax.add_feature(cfeature.LAND,
                                color = self.land_color, 
                                alpha = self.land_alpha, 
                                zorder = 0)
            
        self.ax.coastlines(linewidth = self.lw_coast, zorder=2)
            
    def limits(self):

        self.lon_extents, self.lat_extents, _ = self.ax.projection.transform_points(
                                                        self.plot_projection,
                                                        np.array(self.lon_extents),
                                                        np.array(self.lat_extents)).T

        self.ax.set_extent([*self.lon_extents,
                            *self.lat_extents])
            
    
    def lines(self):
        
        gl = self.ax.gridlines(crs = self.plot_projection, 
                               linewidth = self.lw_lines,
                               color = self.color_lines, 
                               linestyle = self.ls_lines, 
                               draw_labels = True, 
                               x_inline = False, 
                               y_inline = False, 
                               zorder = 5)


        if 'right' not in self.label_lines: 
            gl.right_labels = False
        if 'bottom' not in self.label_lines:
            gl.bottom_labels = False
        if 'top' not in self.label_lines:
            gl.top_labels = False
        if 'left' not in self.label_lines:
            gl.left_labels = False

        self.grid_lines = gl


    def ticks(self):

        ylocator = LatitudeLocator(prune = 'both', 
                                   nbins = self.nticks_y, 
                                   integer = self.integery)
        
        xlocator = LongitudeLocator(prune = 'both', 
                                    nbins = self.nticks_x, 
                                    integer = self.integerx)
        
        self.grid_lines.ylocator = ylocator
        self.grid_lines.xlocator = xlocator

        self.grid_lines.xlabel_style = {'size': self.fs_ticks}
        self.grid_lines.ylabel_style = {'size': self.fs_ticks}

    
    def create(self):

        super().decoration()
        
        self.updates()
        
        self.features()
        self.limits()
        self.lines()
        self.ticks()

        return self
    

@dataclass(kw_only = True)
class boxplot(base_001):

    fs_ticks: float = 9
    
    def create(self):

        super().decoration()

        self.ax.tick_params(axis = 'both',
                            which = 'major',
                            labelsize = self.fs_ticks)
        
        self.ax.yaxis.get_offset_text().set_fontsize(self.fs_ticks)

        return self
            

@dataclass(kw_only = True)
class xy_numeric(base_001):
   
    xs: np.ndarray | pd.DatetimeIndex
    ys: np.ndarray | pd.Series

    fs_ticks: float = 10.0
    ticks_y: bool = True
    ticks_x: bool = True
    nticks_y: int = 5
    nticks_x: int = 5
    integery: bool = False
    integerx: bool = False
    invertx: bool = False
    inverty: bool = False

    x_limits: None | list[float] = None
    y_limits: None | list[float] = None
    y_ticks_rel_b: float = 0.1
    x_ticks_rel_b: float = 0.1

    def ticks(self):

        xlocator = MaxNLocator(prune = 'both', 
                               nbins = self.nticks_x, 
                               integer = self.integerx)
        
        ylocator = MaxNLocator(prune = 'both', 
                               nbins = self.nticks_y, 
                               integer = self.integery)

        self.ax.yaxis.set_major_locator(ylocator)
        self.ax.xaxis.set_major_locator(xlocator)

        if not self.ticks_x: self.ax.set_xticklabels([])
        if not self.ticks_y: self.ax.set_yticklabels([])

        self.ax.tick_params(axis = 'both', 
                            which = 'major', 
                            labelsize = self.fs_ticks)


    def limits(self):

        if self.x_limits is None:
            
            tsmax = np.nanmax(self.xs) 
            tsmin = np.nanmin(self.xs)

            tsabs = np.max(np.abs([tsmax, tsmin]))

            tsmax_rel = tsmax + tsabs * self.x_ticks_rel_b
            tsmin_rel = tsmin - tsabs * self.x_ticks_rel_b

            self.ax.set_ylim(tsmin_rel, tsmax_rel)

        else:

            self.ax.set_ylim(self.x_limits[0], 
                             self.x_limits[1])
            
        if self.y_limits is None:
            
            tsmax = np.nanmax(self.ys) 
            tsmin = np.nanmin(self.ys)

            tsabs = np.max(np.abs([tsmax, tsmin]))

            tsmax_rel = tsmax + tsabs * self.y_ticks_rel_b
            tsmin_rel = tsmin - tsabs * self.y_ticks_rel_b

            self.ax.set_ylim(tsmin_rel, tsmax_rel)

        else:

            self.ax.set_ylim(self.y_limits[0], 
                             self.y_limits[1])
        
        if self.invertx: plt.gca().invert_xaxis()
        if self.inverty: plt.gca().invert_yaxis()
    
    def create(self):

        super().decoration()
        
        self.ticks()
        self.limits()

        return self



def apply(array, axes, method: str):

    """
    Aggregate with the method along the axis
    with the method (e.g. mean, max, mean, variance).
    The method has to be a numpy attribute.
    The given axis dimension(s) will be shrunk.
    """

    import numpy as np

    # find the method in numpy attributes
    method_np           = getattr(np, method)

    # apply the method along given axis
    array_agg           = method_np(array, axis = axes)

    return array_agg


def select_o_apply_d(array, array_sel_4d = None, method_sel = 'argmax', 
                     axis_sel: int = 0, axis_sel_shrink: int = 0, 
                     method_agg_4d: str = 'mean', method_agg_time: str ='mean'):
    
    from my_gridded.aggregate import apply
    from my_gridded.dimensions import select_by_other, select_apply_4d

    var_array_3d                = select_apply_4d(array, sel_func = select_by_other,
                                                    sel_func_args = {'array_sel': array_sel_4d,
                                                        'method': method_sel, 'axis_select': axis_sel,
                                                        'axis_shrink': axis_sel_shrink},
                                                    apply_args = {'axes': axis_sel_shrink,
                                                        'method': method_agg_4d})
    
    var_array_2d                = apply(var_array_3d, axes = 0, method = method_agg_time)

    return var_array_2d
    





import os
import pandas as pd
from glob import glob

def open_multiple_csv(files: os.PathLike, 
                      args: dict = {}, 
                      names = []):

    dfs                         = {}

    #if isinstance(files)
        
    for i_f in range(len(files)):

        name                    = names[i_f]
        file_name               = files[i_f]
        file                    = glob(file_name)[0]
        df                      = pd.read_csv(file, **args)

        dfs.update({name: df})

    return dfs


def open_csv(file_str: os.PathLike, 
             **args) -> pd.DataFrame:

    import pandas as pd
    from glob import glob

    file                       = glob(file_str)[0]

    df                          = pd.read_csv(file, **args)

    return df




import numpy as np
from my_.series.convert import tab_to_array


def rmse(obs, sim, decimals=2):

    """
    Root mean square error 
    Use this to compare model to observation series
    """    

    obs = tab_to_array(obs)
    sim = tab_to_array(sim)
    
    error = sim - obs

    if np.all(np.isnan(error)): return np.nan

    squared_error = error**2

    sum_squared_error = np.nansum(squared_error)

    n = np.count_nonzero(~np.isnan(error))

    mean_squared_error = sum_squared_error / n

    rmse = mean_squared_error**0.5

    rmse_rounded = np.around(rmse, decimals)
    
    return rmse_rounded


def pbias(obs, sim, decimals: int = 2):

    obs = tab_to_array(obs)
    sim = tab_to_array(sim)

    bias = sim - obs

    if np.all(np.isnan(bias)): return np.nan

    sum_bias = np.nansum(bias)

    sum_obs = np.nansum(obs)

    rel_bias = sum_bias / sum_obs
    
    percent_bias = rel_bias * 100

    percent_bias_rounded = np.around(percent_bias, decimals)

    return percent_bias_rounded


def r(obs, sim, decimals: int = 2):

    from scipy.stats.stats import pearsonr

    obs = tab_to_array(obs)
    sim = tab_to_array(sim)

    mask = ~np.isnan(obs) & ~np.isnan(sim)
    
    if np.count_nonzero(mask) <= 5: return np.nan

    obs_masked = obs[mask]
    sim_masked = sim[mask]

    r, p = pearsonr(obs_masked, sim_masked)

    r_rounded, p_rounded = np.around(r, decimals), np.around(p, decimals)

    return r_rounded


from dataclasses import dataclass
import importlib
import os
import numpy as np
from typing import Any
import pandas as pd
import pint
import xarray as xr
import pint_xarray
from glob import glob
from datarie.netcdf import nc_open, variables_to_array
from datarie.handy import check_file_exists

@dataclass
class gridded_data:
    
    name: str
    version: tuple[int]
    path: str
    type_file: str
    year_start: int
    month_start: int
    year_end: int
    month_end: int
    resolution_time: str
    leapday: bool
    grid: str
    variables: list[str]
    variable_names: dict[str, Any]
    variable_dimensions: dict[str, Any]
    variable_units: dict[str, Any]
    mask_value: None | float | int = None
    

    def index_time(self,
                   y0: int | None = None,
                   y1: int | None = None,
                   month_start: int | None = None) -> pd. Series:
        
        from datarie.time import index
        
        if y0 is None: y0 = self.year_start
        elif y0 < self.year_start: y0 = self.year_start
        
        if y1 is None: y1 = self.year_end
        elif y1 > self.year_end: y1 = self.year_end

        if month_start is None: month_start = self.month_start

        time_index = index(y0 = y0, 
                           y1 = y1, 
                           t_res = self.resolution_time,
                           leapday = self.leapday)
    
        return time_index.loc[f'{y0}-{month_start}-01':]
    

    def present_files(self,
                      y0: int | None,
                      y1: int | None):
        
        if y0 is None: y0 = self.year_start
        elif y0 < self.year_start: y0 = self.year_start
        
        if y1 is None: y1 = self.year_end
        elif y1 > self.year_end: y1 = self.year_end
    
        files = [f'{self.path}/{y}.nc' for y in np.arange(y0, y1 +1)]

        return files

    def present_variables(self, 
                          load_variables: list[str]):
        
        variables_present = [v for v in load_variables if v in self.variables]

        return variables_present
    

    def present_source_variables(self,
                                 load_variables: list[str]):
    
        variables_present = self.present_variables(load_variables)
            
        source_variables = [self.variable_names[v] for v in variables_present]
        
        return source_variables
        
    
    def get_values(self,
                   variables_: str | list[str] | None,
                   y0: int | None = None,
                   y1: int | None = None,
                   dtype: str = 'float32'):
        
        if isinstance(variables_, str): 
            variables_ = [variables_]

        if y0 is None: y0 = self.year_start
        elif y0 < self.year_start: y0 = self.year_start
        
        if y1 is None: y1 = self.year_end
        elif y1 > self.year_end: y1 = self.year_end

        if variables_ is None: variables_ = self.variables

        print(f'\nLoading value data for:')
        print(f'{variables_} from year {y0} to year {y1}...\n')

        files = [f'{self.path}/{y}.nc'
                for y in range(y0, y1 + 1)]

        data = nc_open(files)

        present_vars = self.present_variables(variables_)

        present_src_vars = self.present_source_variables(variables_)
        
        values = variables_to_array(data, 
                                    present_src_vars,
                                    dtype = dtype,
                                    mask_value = self.mask_value)

        return {v: values[i] for i, v in enumerate(present_vars)}
    

    def get_variables(self, 
                      load_variables: list[str] | None = None,
                      y0: int | None = None,
                      y1: int | None = None,
                      rename: bool = False):
        
        if y0 is None: y0 = self.year_start
        elif y0 < self.year_start: y0 = self.year_start
        
        if y1 is None: y1 = self.year_end
        elif y1 > self.year_end: y1 = self.year_end

        print(f'\nLoading data for:')
        print(f'{load_variables} from year {y0} to year {y1}...\n')

        if load_variables is None: load_variables = self.variables
        
        present_vars = self.present_variables(load_variables)

        type_module = check_module_var('my_.files',
                                        self.type_file)

        files = [f'{self.path}/{y}.{type_module}.nc'
                for y in range(y0, y1 + 1)]
        
        data = xr.open_mfdataset(files) if len(files) > 1 else xr.open_dataset(files[0])

        name_dict = {v: k for k, v in self.variable_names.items() if k in present_vars}

        if rename: 
            data = data.rename(name_dict)[present_vars]
                           
        return data
    
    def apply_landmask(self,
                       variables: dict | xr.Dataset,
                       landmask: np.ndarray):

        if isinstance(variables, xr.Dataset):
            return variables.where(landmask)
        
        masked_dict = {k: np.where(landmask, 
                                   v, 
                                   np.nan) 
                       for k, v in variables.items()}
        
        return masked_dict


    def convert_units(self,
                      values: dict[str, np.ndarray],
                      dst_units: dict):
        
        ureg = pint.UnitRegistry()

        values_out = {} if isinstance(values, dict) else values.copy()

        for v, array in values.items():

            src_unit = self.variable_units[v]
            dst_unit = dst_units[v]
        
            print(f'\nConverting {v} units from {src_unit} to {dst_unit}...\n')

            if isinstance(array, np.ndarray):
                
                Q_ = ureg.Quantity
                
                values_ = Q_(array, src_unit)
                
                values_dst = values_.to(dst_unit)

                values_out[v] = values_dst.magnitude

            #elif isinstance(array, xr.DataArray):

            #    array.lat.attrs['units'] = 'degree'
            #    array.lon.attrs['units'] = 'degree'

            #    values = array.pint.quantify({f'{v}': src_unit})

            #    values_dst = values.pint.to(dst_unit)

            #    values_out[v] = values_dst.pint.dequantify()

        return values_out


    def regrid(self):
        ...

    def resample(self,
                 dataset: xr.DataArray | xr.Dataset,
                 dst_time_res: str,
                 method: str | list,
                 var_subset: list | None = None,
                 **kwargs):

        from datarie.time import xr_resample
        
        if var_subset is not None: dataset = dataset[var_subset]

        if isinstance(method, list):

            out_ds = xr.Dataset()
            
            for m in method:

                ds = xr_resample(dataset, dst_time_res, m, **kwargs)
    
                new_names = {n: f'{n}_{m}' for n in ds.keys()}
    
                ds_new = ds.rename(new_names)
        
                out_ds.merge(ds_new)

        elif isinstance(method, str):

            out_ds = xr_resample(dataset, dst_time_res, method, **kwargs)
    
        return out_ds
    
    def save(self,
             dataset: xr.Dataset | xr.DataArray,
             out_file: os.PathLike):
        
        if check_file_exists(f'{out_file}'): return
        
        dataset.to_netcdf(f'{out_file}',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])



@dataclass
class grid:

    name: str
    version: tuple[str]
    path_file: os.PathLike
    name_file: str
    type_file: str
    name_latitude: str
    name_longitude: str
    name_landmask: str
    lon_extents: list[float]
    lat_extents: list[float]

    def load_coordinates(self) -> list[np.ndarray]:

        file = f'{self.path_file}/{self.name_file}'

        data = nc_open(file)

        values = variables_to_array(data, 
                                    [self.name_latitude, 
                                     self.name_longitude])

        return values

    def load_landmask(self) -> np.ndarray:

        file = f'{self.path_file}/{self.name_file}'

        data = nc_open(file)

        values = variables_to_array(data, 
                                    [self.name_landmask])

        return values[0]


    def points(self):

        lat, lon = self.load_coordinates()

        if (lat.ndim == 1) & (lon.ndim == 1):
            
            stacked_coords = np.array(np.meshgrid(lat, 
                                                  lon))
        
            list_points = stacked_coords.T.reshape(-1, 2)

        elif (lat.ndim == 2) & (lon.ndim == 2):

            stacked_coords = np.dstack([lat, lon])
    
            list_points = stacked_coords.reshape(-1, 2)
    
        return list_points


    def shape(self):

        lat, lon = self.load_coordinates()

        if (lat.ndim == 1) & (lon.ndim == 1):
            
            shape = (*lat.shape, *lon.shape)
    
        elif (lat.ndim == 2) & (lon.ndim == 2):

            shape = lat.shape

        return shape
    
    
    def save(self,
             dataset: xr.Dataset | xr.DataArray,
             out_file: os.PathLike):
        
        dataset.to_netcdf(f'{out_file}',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])


def check_module_var(module: str, 
                     var: str) -> dict:

    try:
    
        mod = importlib.import_module(module)
            
        print(f'\nModule {module} imported successfully.\n')

        return getattr(mod, var)
    
    except ModuleNotFoundError: 
        
        print(f'\nModule {module} is not yet supported. Continue...\n')

        return dict()





def src_var_cells_df(name_case: str, 
                     sources: list, 
                     variables: list, 
                     file_format: str = 'csv',
                     path_stations: str = 'user_in/csv/', 
                     file_stations: str = 'stations.csv',
                     path_out: str = 'out/',
                     year_start: int = 1995, 
                     year_end: int = 2018):

    print(f'\nIterate over gridded source data:\n{*sources,}\n')

    from glob import glob
    from itertools import product
    import importlib

    from my_.resources.sources import query_variables, query_grids, available_variables
    from my_.files.handy import yearly_or_monthly_files, save_df, create_dirs
    from my_.files.netcdf import nc_open, variables_to_array
    from my_.files.csv import open_csv
    from my_.grids.dimensions import grid_to_points
    from my_.series.aggregate import concat
    from my_.series.group import single_src_var_X_index
    from my_.series.time import index
    from my_.resources.units import transform
    from my_.grids.extract import grid_to_cell_df

    for src in sources:

        #importlib.import_module(f'my_.data_{src}')
        
        dir_out                 = f'{path_out}/{name_case}/{file_format}/'
        dir_csv                 = f'{path_out}/{name_case}/csv/'

        create_dirs([dir_out, dir_csv])
        
        file_out                = f'{dir_out}/Extracted_{src}.{file_format}'
        file_out_info           = f'{dir_csv}/Cells_{src}.csv'

        if glob(file_out): print(f'Output file for {name_case} - {src} is already available.\n'); continue
        
        print(f'Output file for {src} is not available and will be created.\n')

        df_stations             = open_csv(f'{path_stations}/{file_stations}')
        df_cells_info           = df_stations.copy()

        names_stations          = df_stations['name']
        lat_stations            = df_stations['lat']
        lon_stations            = df_stations['lon']
        landcover               = df_stations['landcover']

        path                    = query_variables(src, 'path')
        freq_files              = query_variables(src, 'freq_files')
        time_step               = query_variables(src, 'time_step')
        leapday                 = query_variables(src, 'leap_day')
        grid                    = query_variables(src, 'grid')
        
        file_grid               = query_grids(grid, 'file')
        path_grid               = query_grids(grid, 'path')
        grid_lat                = query_grids(grid, 'var_lat')
        grid_lon                = query_grids(grid, 'var_lon')

        vars_avail, vars_src    = available_variables(src, variables)

        if not vars_src: print(f'Variable(s) not available for {src}. Continue...\n'); continue

        files                   = yearly_or_monthly_files(freq_files, path, year_start, year_end)

        print('Load data...\n')
        data                    = nc_open(files)
        arrays_list             = variables_to_array(data, variables = vars_src, dtype = 'float32')

        data_geo                = nc_open(f'{path_grid}/{file_grid}')
        
        [lat2d, lon2d]          = variables_to_array(data_geo, variables = [grid_lat, grid_lon], dtype ='float32')

        list_points             = grid_to_points(lat2d, lon2d)

        time                    = index(y0 = year_start, 
                                        y1 = year_end, 
                                        t_res = time_step, 
                                        leapday = leapday)

        iterations              = product(range(len(names_stations)), range(len(vars_avail)))

        print('Iterate stations and variables...\n')

        list_ts                 = []

        for i_station, i_var in iterations:

            name_station        = names_stations[i_station]
            lc                  = landcover[i_station]
            lat_yx              = lat_stations[i_station]
            lon_yx              = lon_stations[i_station]
            var                 = vars_avail[i_var]
            array               = arrays_list[i_var]

            print(f"Extract site {names_stations[i_station]}, variable {vars_avail[i_var]}")
            print(f'Location: {lat_yx} latitude, {lon_yx} longitude.\n')

            column              = single_src_var_X_index(array, src, var, lc, name_station)
            
            idxs, coords, ts    = grid_to_cell_df(array = array,
                                                  points = list_points, 
                                                  shape = lat2d.shape, column = column,
                                                  latitude = lat_yx, longitude = lon_yx,
                                                  index = time)
            
            ts_ux               = transform(ts, src, var)

            list_ts.append(ts_ux)

            df_cells_info.loc[i_station,['lat_cell', 'lon_cell']] = coords
            df_cells_info.loc[i_station,['i_lat_cell', 'i_lon_cell']] = idxs
        
        df_out = concat(list_ts)

        save_df(df_cells_info, file_out_info, 'csv')
        save_df(df_out, file_out, file_format)
    
    print('Done. Extracted cell data for all sources.\n')


def cell_static_info(name_case: str, sources = [], path_stations: str = 'user_in/csv/', file_stations: str = 'stations.csv'):

    """
    Take cell info dataframe from above and include 
    Additional static information on that cell
    """

    from glob import glob
    from itertools import product
    
    from my_.resources.sources import query_static, query_grids
    from my_.grids.dimensions import grid_to_points
    from my_.files.netcdf import nc_open, variables_to_array
    from my_.files.csv import open_csv
    from my_.grids.extract import closest_cell, cell
    from my_.files.handy import save_df
    
    for src in sources:

        file_out                = f'out/{name_case}/csv/Static_data_{src}.csv'

        if glob(file_out): print(f'Static data output file for {src} is already available.\n'); continue

        df_stations             = open_csv(f'{path_stations}/{file_stations}')
        df_out                  = df_stations.copy()

        names_stations          = df_stations['name']
        lat_stations            = df_stations['lat']
        lon_stations            = df_stations['lon']

        path                    = query_static(src, 'path')
        file                    = query_static(src, 'file')
        grid                    = query_static(src, 'grid')
        variables               = query_static(src, 'static_vars')
        
        file_grid               = query_grids(grid, 'file')
        path_grid               = query_grids(grid, 'path')
        grid_lat                = query_grids(grid, 'var_lat')
        grid_lon                = query_grids(grid, 'var_lon')
        
        data_geo                = nc_open(f"{path_grid}/{file_grid}")
        
        [lat2d, lon2d]          = variables_to_array(data_geo, variables = [grid_lat, grid_lon])

        data                    = nc_open(f'{path}/{file}')
        arrays_list             = variables_to_array(data, variables = variables)

        # List of lat lon grid cell coordinates
        list_points             = grid_to_points(lat2d, lon2d)

        iterations              = product(range(len(names_stations)), range(len(variables)))

        for i_station, i_var in iterations:

            name_station        = names_stations[i_station]
            var                 = variables[i_var]
            lat_yx              = lat_stations[i_station]
            lon_yx              = lon_stations[i_station]
            array               = arrays_list[i_var]

            print(f'Extract site {name_station}, static data for {src} variable {var}')
            print(f'Location: {lat_yx} latitude, {lon_yx} longitude.\n')
            
            grid_shape          = lat2d.shape
            ndim                = array.ndim
            length              = array.shape[0]
            
            name_out            = f"{var}"
            names_out           = name_out if ndim == 2 else [f'{name_out}_{i}' for i in range(length)]

            [y, x], coords_cell = closest_cell([lat_yx, lon_yx], list_points, grid_shape)

            cell_station        = cell(array, y = y, x = x)

            df_out.loc[i_station, ['lat_cell', 'lon_cell']] = coords_cell
            df_out.loc[i_station, ['i_lat_cell', 'i_lon_cell']] = y, x
            df_out.loc[i_station, names_out] = cell_station

        save_df(df_out, file_out, 'csv')



if __name__ == '__main__':

    from argparse import ArgumentParser

    parser                      = ArgumentParser()

    parser.add_argument('--name', '-n', help='Name for your work', type=str)
    parser.add_argument('--sources', '-s', nargs='+', help='List of source names', type=str)
    parser.add_argument('--variables', '-v', nargs='+', help='List of variables', type=str)
    parser.add_argument('--path_stations', '-ps', help='Path to the stations file', default='user_in/csv/', type=str)
    parser.add_argument('--file_stations', '-fs', help='File name of the stations info', default='stations.csv', type=str)
    parser.add_argument('--file_format', '-ff', help='The file format ending of the output file', default='csv', type=str)
    parser.add_argument('--include_static', '-is', help='Include static information from source?', default=False, type=bool)
    parser.add_argument('--static_sources', '-ss', nargs='+', help='Static sources for extract', default='', type=str)
    parser.add_argument('--year_start', '-y0', help='Start year', type=int)
    parser.add_argument('--year_end', '-y1', help='Stop year', type=int)

    args                        = parser.parse_args()

    src_var_cells_df(name_case = args.name, sources = args.sources, variables = args.variables,
                    path_stations = args.path_stations, file_stations = args.file_stations,
                    file_format = args.file_format, year_start = args.year_start, year_end = args.year_end)
    
    
    if args.include_static:
        
        cell_static_info(args.name, args.static_sources, args.path_stations, args.file_stations)




import numpy as np
import xarray as xr
from cartopy.crs import Projection
from my_.plot.basic import plot
           

regions = {'BI': [(50, -10), (50, 2), (59, 2), (59, -10)],
           'IP': [(36, -10), (36, 3), (44, 3), (44, -10)],
           'FR': [(44, -5), (44, 5), (50, 5), (50, -5)],
           'ME': [(48, 2), (48, 16), (55, 16), (55, 2)],
           'SC': [(55, 5), (55, 30), (70, 30), (70, 5)],
           'AL': [(44, 5), (44, 15), (48, 15), (48, 5)],
           'MD': [(36, 3), (36, 25), (44, 25), (44, 3)],
           'EA': [(44, 16), (44, 30), (55, 30), (55, 16)],}


def plot_regions(ax, 
                    color: str = 'k',
                    lw: float = 1.3,
                    projection: None | Projection = None,
                    alpha: float = 0.8):
    
    for reg, points in regions.items():
        
        for i, v in enumerate(points):

            next = i + 1 if i < 3 else 0
        
            lats = np.linspace(points[i][0], points[next][0], 1000)
            lons = np.linspace(points[i][1], points[next][1], 1000)

            plot(ax, lons, lats,
                    colors = color,
                    lw = lw,
                    projection = projection,
                    alpha = alpha)


def mask_prudence(array: np.ndarray | xr.DataArray,
                  lat: np.ndarray | xr.DataArray,
                  lon: np.ndarray | xr.DataArray,
                  sel_regions: str | list[str] | None):

    if sel_regions is None: return array

    shape = array.shape[-2:]

    mask = np.zeros(shape)

    for r in sel_regions:

        lat_max = regions[r][2][0]
        lat_min = regions[r][0][0]
        lon_max = regions[r][2][1]
        lon_min = regions[r][0][1]

        mask = np.where(((lat <= lat_max) & 
                        (lat >= lat_min) &
                        (lon <= lon_max) &
                        (lon >= lon_min)), 1, mask)
        
    if isinstance(array, xr.DataArray): 

        array_out = array.where(mask)

    return array_out


import pandas as pd

def check_file_exists(str_file):

    from glob import glob

    if len(glob(str_file)) > 0:
        
        print(f'\nOutput file(s) {str_file} exists. Return.\n')
        
        return True
    
    else:
        
        print(f'\nOutput file(s) {str_file} does not exist.')
        print('It will be created now...\n')
        
        return False


def yearly_files(path, y0: int = 1995, y1: int = 2018):

    files = [f"{path}/{str(y)}.nc" for y in range(y0, y1+1)]

    return files


def monthly_files(path, y0: int = 1995, y1: int = 2018):

    files = [f"{path}/{str(y)}-{str(m).zfill(2)}.nc" for y in range(y0, y1+1) for m in range(1,13)]

    return files


def yearly_or_monthly_files(freq_files, path, y0, y1):

    if freq_files == 'monthly':

        files = monthly_files(path, y0, y1)

    elif freq_files == 'yearly':

        files = yearly_files(path, y0, y1)

    return files


def save_df(df: pd.DataFrame, 
            name: str, 
            format: str = 'csv', 
            mode: str = 'w'):

    if format == 'csv':
            
        df.to_csv(name)
        
    elif format == 'parquet':
           
        df.to_parquet(name)


def open_csv_or_parquet(file: str, 
                        csv_args = {'index_col': 0}, 
                        parquet_args = {}):

    from datarie.csv import open_csv
    from datarie.parquet import open_parquet

    ending = file.split('.')[-1]
    
    open_method = open_csv if ending == 'csv' else open_parquet

    args = csv_args if ending == 'csv' else parquet_args

    df = open_method(file, **args)

    return df

    
def create_dirs(dirs: str):

    from pathlib import Path

    if isinstance(dirs, str): 
        
        dirs_= [dirs]

    else: 

        dirs_ = dirs
        
    for dir in dirs_:

        Path(dir).mkdir(parents = True, exist_ok = True)






import netCDF4 as nc
import numpy as np
import psutil
from typing import Literal

def nc_open(str_files: str | list,
            **kwargs):

    from glob import glob
    
    if isinstance(str_files, str):

        files = glob(str_files)

        n_files = len(files)

        if n_files == 1: return nc.Dataset(files[0], **kwargs)

        data = nc.MFDataset(sorted(files), **kwargs)

    elif isinstance(str_files, list):

        data = nc.MFDataset(sorted(str_files), **kwargs)
        
    return data


def variable_to_array(data: nc.Dataset, 
                      variable: str, 
                      stack_axis: int = 1,
                      dtype: str = 'float32',
                      mask_value: None | float | int = None):

    print(f'\nLoad netcdf variable {variable} to memory.')
    print(f'Available memory [GB]: {psutil.virtual_memory()[4] / 10**9}...\n')

    if isinstance(variable, list):

        arrays = [data.variables[v][:] 
                  for v in variable]

        array = np.stack(arrays, 
                         axis = stack_axis)
              
    elif isinstance(variable, str):

        netcdf_variable = data.variables[variable]

        array = netcdf_variable[:]

    np_dtype = getattr(np, dtype)
    
    array_out = array.astype(np_dtype)

    if mask_value is not None:

        array_out = np.ma.masked_values(array_out, mask_value)

    if isinstance(array_out, np.ma.masked_array): return array_out.filled(np.nan)
    else: return array_out
    

def variables_to_array(data: nc.Dataset, 
                       variables: list[str],
                       stack_axis: int = 1, 
                       dtype: str = 'float64',
                       mask_value: None | float | int = None):
    
    arrays_dtype = [variable_to_array(data, 
                                      v, 
                                      stack_axis,
                                      dtype,
                                      mask_value) for v in variables]

    return arrays_dtype


def variables_to_dict(data: nc.Dataset, 
                       variables: list[str],
                       stack_axis: int = 1, 
                       dtype: str = 'float64',
                       mask_value: None | float | int = None):
    
    if isinstance(variables, str): 
        variables = [variables]
    
    arrays_dtype = {v: variable_to_array(data, 
                                         v, 
                                         stack_axis,
                                         dtype,
                                         mask_value) for v in variables}

    return arrays_dtype

    
def netcdf_time(data: nc.Dataset,
                name_time: str = 'time',
                only_use_python_datetimes: bool = False,
                **kwargs):
    
    ## BAUSTELLE

    var = data.variables[name_time]
    units = var.units
    calendar = var.calendar

    dtime = nc.num2date(var[:], 
                        units = units, 
                        calendar = calendar,
                        only_use_python_datetimes = only_use_python_datetimes,
                        **kwargs)

    return dtime


def nc_out(fname: str,
           dimensions: dict,
           variables: dict,
           attributes: dict = {}, 
           format: Literal['NETCDF4_CLASSIC', 
                           'NETCDF3_CLASSIC'] = 'NETCDF4_CLASSIC',
           ) -> nc.Dataset:


    # dims = {'lat': {'dtype': np.float64, 'dims': [lat, lon], 'size': 16, 'attrs': {'units': 'degrees east'}}}
    # vars = {'temp: {'dtype': np.float64, 'dims':  [lat, lon], 'attrs': {'units': 'degrees east'}}}

    print(f'\nCreating netCDF output to: {fname}...\n')

    nc_out = nc.Dataset(fname, 
                        mode = 'w',
                        format = format)
    
    for k, v in attributes.items():

        nc_out.setncattr(k, v)
    
    for dim, dimc in dimensions.items():

        dim_nc = nc_out.createDimension(dim, dimc['size'])

        dim_var = nc_out.createVariable(dim, dimc['dtype'], dimc['dims'])

        for k, v in dimc['attrs'].items():
        
            dim_var.setncattr(k, v)
    
    for var, varc in variables.items():
        
        var_nc = nc_out.createVariable(var, varc['dtype'], varc['dims'])

        for k, v in varc['attrs'].items():
        
            var_nc.setncattr(k, v)

    return nc_out



"""
A python script to extract netcdf grid-cell data from any grid
Stations information in csv file
The out put time series can be in parquet or csv
"""
import numpy as np
import pandas as pd


def closest_cells(coords_points: np.ndarray, 
                  coords_cells: np.ndarray, 
                  shape: tuple[int]) -> np.ndarray:

    cells = [closest_cell(coords_points[i], 
                          coords_cells, 
                          shape)[0] 
                          for i in range(len(coords_points))]

    return np.array(cells)


def closest_cell(coords_point: np.ndarray, 
                 coords_cells: np.ndarray, 
                 shape: tuple[int]) -> tuple[np.ndarray, np.ndarray]:
    
    """
    Find the closest coordinate point.
    Given a point in 2d coordinates,
    and a list of points find the node.
    """

    if not isinstance(coords_point, np.ndarray): np.array(coords_point)
   
    deltas          = np.subtract(coords_cells, coords_point)

    dist            = np.einsum('ij,ij->i', deltas, deltas)

    closest_i       = np.nanargmin(dist)

    closest_coords  = coords_cells[closest_i, :]

    closest_cell    = np.unravel_index(closest_i, shape)

    return closest_cell, closest_coords


def grid_to_cell_df(array: np.ndarray, 
                    points: np.ndarray, 
                    shape: tuple[int], 
                    column: pd.Index | pd.MultiIndex, 
                    latitude: float, 
                    longitude: float,
                    index: pd.Index | pd.Series,
                    dtype = str) -> tuple[np.ndarray, np.ndarray, pd.Series]:

    """
    Get data from 3d or 4d array
    for one cell along time dimension
    Grid contains the lat lon points 
    lat lon are dims -2 and -1
    time is dim 0
    """

    from my_.series.convert import cell_to_df


    coords_point                = (latitude, longitude)

    indices_cell, coords_cell   = closest_cell(coords_point, points, shape)

    y_cell, x_cell              = [*indices_cell]

    array_cell                  = cell(array, y_cell, x_cell)

    timeseries_cell             = cell_to_df(array_cell, 
                                             column, 
                                             index,
                                             dtype).astype('float') 

    return indices_cell, coords_cell, timeseries_cell


def cell(array: np.ndarray, 
         y: int, 
         x: int, 
         y_dim: int = -2, 
         x_dim: int = -1) -> np.ndarray:

    """
    Get data from 3d or 4d array
    for one cell along time dimension
    Grid contains the lat lon points 
    lat lon are dims -2 and -1
    time is dim 0
    """
    
    #ndim                        = array.ndim
    #print(ndim)
    #
    #indx                        = [slice(None)] * ndim
    #indx[y_dim]                 = slice(y)
    #indx[x_dim]                 = slice(x)

    cell                        = array[..., y, x]
    
    return cell
        

def cells(array: np.ndarray, 
          ys: list[float] | np.ndarray, 
          xs: list[float] | np.ndarray, 
          y_dim: int = -2, 
          x_dim: int = -1):

    cells                       = [cell(array, ys[i], xs[i]) for i in range(len(ys))]

    return cells


def sites(source: str,
          variables: list[str],
          sites: pd.DataFrame,
          year_start: int,
          year_end: int,
          file_out: str | None = None,
          type_out: str = 'csv') -> pd.DataFrame:
    
    from my_.data.templates import gridded_data, grid, check_data_module
    from my_.series.group import single_multiindex
    from my_.series.aggregate import concat
    from my_.files.handy import save_df, check_file_exists
    from itertools import product

    if file_out is None: file_out = f'out'

    if check_file_exists(f'{file_out}.{type_out}'): return

    data_module = check_data_module('my_.data',
                                    source)
    if not data_module: return

    data = gridded_data(**data_module)
    
    data_variables = data.get_values(variables,
                                     y0 = year_start,
                                     y1 = year_end)
    
    grid_module = check_data_module(f'my_.data', 'grids')
    grid_attributes = getattr(grid_module, data.grid)

    data_grid = grid(**grid_attributes)
    points = data_grid.points()
    shape = data_grid.shape()
    index_time = data.index_time(year_start,
                                 year_end)

    list_ts = []

    n_sites = len(sites['name'])

    print(f'\nBegin loop over {n_sites} sites...\n')
    
    sites_out = sites.copy()

    for site, var in product(sites.index, variables):

        print(f'Get {var} data from location of site {site}...\n')
        
        site_latitude = sites['latitude'][site]
        site_longitude = sites['longitude'][site]
        site_name = sites['name'][site]
        site_type = sites['ecosystem_type'][site]

        units = data.variable_units[var]

        var_array = data_variables[var]

        if var_array.ndim == 3:
        
            indices = {'Variable': var_array, 
                       'Source': source, 
                       'Variable': var,
                       'Unit': units, 
                       'Ecosystem type': site_type, 
                       'Site': site_name}
            
        elif var_array.ndim == 4:
                       
            layers = np.arange(var_array.shape[1]).astype(str).tolist()
        
            indices = {'Variable': var_array, 
                       'Source': source, 
                       'Variable': var,
                       'Unit': units, 
                       'Ecosystem type': site_type, 
                       'Site': site_name,
                       'Layer': layers}


        column = single_multiindex(indices)
                        
        idxs, coords, ts = grid_to_cell_df(data_variables[var], 
                                           points, 
                                           shape, 
                                           column, 
                                           site_latitude, 
                                           site_longitude,
                                           index_time)

        list_ts.append(ts)
                    
        sites_out.loc[site, 'i_lat'] = idxs[0]
        sites_out.loc[site, 'i_lon'] = idxs[1]
        sites_out.loc[site, 'cell_lat'] = coords[0]
        sites_out.loc[site, 'cell_lon'] = coords[1]
    
    df_out = concat(list_ts, 
                    sort = False)

    save_df(sites_out,
            f'deims/sites_out_{source}.csv')
    
    save_df(df_out, 
            f'{file_out}.{type_out}', 
            type_out)


    
    



from cartopy.crs import Projection
import matplotlib.pyplot as plt
import numpy as np


def colormesh(ax, x, y, array, cmap: str = 'coolwarm_r', v0: float = None, v1: float = None, projection = None):

    artist = ax.pcolormesh(x, y, array, cmap = cmap, vmin = v0, vmax = v1, transform = projection, zorder=0)
    
    return artist


def scatter(ax: plt.Axes, 
            xs: np.ndarray, 
            ys: np.ndarray, 
            marker: str | None = None, 
            sizes_marker: int = 50, 
            colors_marker: str = 'black', 
            edgecolor: str | list | None = None,
            face: bool = True, 
            alpha: float = 0.7, 
            projection: Projection | None = None, 
            zorder: int = 5):

    if isinstance(sizes_marker, int): 
        sizes_marker = [sizes_marker] * len(xs)

    if isinstance(colors_marker, str): 
        colors_marker = [colors_marker] * len(xs)

    if isinstance(edgecolor, str):
        edgecolor = [edgecolor] * len(xs)

    if projection == None: 
        projection = ax.transData
    
    artist = ax.scatter(xs, 
                        ys, 
                        marker = marker, 
                        s = sizes_marker, 
                        c = colors_marker,
                        edgecolors = edgecolor,
                        alpha = alpha, 
                        transform = projection, 
                        zorder = zorder)
    
    if not face: artist.set_facecolor("none")

    return artist


def xy(ax, xs, ys, marker, sizes_marker = 50, colors_marker = [], projection = None, 
       axhv_color: str = 'k', axhv_ls: str = '--', axhv_lw: float = 0.7, axhv_alpha: float = 0.8, axhv_dashes: tuple = (4, 4),
       diag_color: str = 'k', diag_ls: str = '--', diag_lw: float = 0.7, diag_alpha: float = 0.8, diag_dashes = (4,4),
       xlabel: str = '', ylabel: str = '', fs_label: int = 9):

    from my_.plot.basic import scatter
    from my_.plot.init_ax import init_xy
    
    init_xy(ax, xs, ys, axhv_color, axhv_ls, axhv_lw, axhv_alpha, axhv_dashes,
       diag_color, diag_ls, diag_lw, diag_alpha, diag_dashes, xlabel, ylabel, fs_label)

    artist = scatter(ax, xs, ys, marker = marker, sizes_markers = sizes_marker, colors_markers = colors_marker, projection = projection)

    return artist


def bar(ax, xs, ys, color, width: float = 0.8, alpha: float = 0.8):

    artist = ax.bar(xs, ys, color = color, width = width, alpha = alpha, zorder = 5)

    return artist


def hist(ax, array, bins = 'auto', density = True,  histtype = 'stepfilled', color= 'dimgray', alpha: float = 0.8, zorder: int = 3):

    artist = ax.hist(array, bins = bins, density = density, histtype = histtype, color = color, alpha = alpha, zorder = zorder)

    return artist


def plot(ax: plt.Axes, 
         xs: np.ndarray, 
         ys: np.ndarray, 
         colors: str | list = 'k', 
         edgecolors: str | list = 'k',
         style: str = '-', 
         lw: float = 1.0, 
         alpha: float = 0.8,
         cycle_direction: int = 0,
         projection: None | Projection = None,
         markersize: float = 1.0,
         marker: str | list = '', 
         fillstyle = 'full',
         zorder = 5):
    
    l = ys.shape[1] if cycle_direction else len(xs)

    if isinstance(lw, float):
        lw = [lw] * l

    if isinstance(markersize, int) \
       or isinstance(markersize, float): 
       markersize = [markersize] * l

    if isinstance(colors, str): 
        colors = [colors] * l

    if len(colors) == 1:
        colors = colors * l

    if projection is None: projection = ax.transData

    ax.set_prop_cycle(color = colors,
                      linewidth = lw,
                      markersize = markersize)

    artist = ax.plot(xs, ys, 
                     transform = projection, 
                     marker = marker, 
                     ls = style,
                     fillstyle = fillstyle,
                     alpha = alpha, 
                     zorder = zorder)

    return artist


def fill(ax, xs, y1s, y2s, colors, alpha: float = 0.4, zorder: int = 2):

    if (y1s.ndim == 2):

        for iy, y1 in enumerate(y1s):
            
            artist = ax.fill_between(xs, y1s[y1], 
                                     y2s.iloc[:,iy], 
                                     color = colors[iy], 
                                     alpha = alpha, 
                                     zorder = zorder)

    return artist


def pie(ax, shares, colors, **kwargs):

    ax.pie(shares, colors = colors, **kwargs)


def error(ax: plt.Axes, 
          xs: np.ndarray, 
          ys: np.ndarray, 
          x_err: np.ndarray | None = None, 
          y_err: np.ndarray | None = None, 
          ecolors: list | str | None = None, 
          elinewidth: str | None = None, 
          capsize: float | None = 0.0, 
          capthick: float | None = None,
          alpha: float = 0.8, 
          zorder: int = 5):
    
    if x_err is None: x_err = [0] * len(y_err)
    if y_err is None: y_err = [0] * len(x_err)
    if isinstance(ecolors, str): ecolors = [ecolors]

    for i in range(len(ecolors)):

        artist = ax.errorbar(xs[i], ys[i], 
                             xerr = x_err[i], yerr = y_err[i], 
                             ecolor = ecolors[i],
                             elinewidth = elinewidth,
                             capsize = capsize, 
                             capthick = capthick,
                             linestyle='', 
                             alpha = alpha, 
                             zorder = zorder)
    
    return artist

def hlines(ax: plt.Axes,
           ys: np.ndarray,
           xmin: np.ndarray,
           xmax: np.ndarray,
           colors: str | list = None,
           linestyle: str = 'solid',
           linewidth: int | float = 2,
           alpha: float = 0.8,
           zorder: int = 5):
    
    artist = ax.hlines(y = ys,
                       xmin = xmin,
                       xmax = xmax,
                       colors = colors,
                       linestyle = linestyle,
                       linewidth = linewidth,
                       alpha = alpha,
                       zorder = zorder)
    
    return artist


def vlines(ax: plt.Axes,
           xs: np.ndarray,
           ymin: np.ndarray,
           ymax: np.ndarray,
           colors: str | list = None,
           linestyle: str = 'solid',
           linewidth: int | float = 2,
           alpha: float = 0.8,
           zorder: int = 5):
    
    artist = ax.axvline(x = xs,
                        ymin = ymin,
                        ymax = ymax,
                        color = colors,
                        linestyle = linestyle,
                        linewidth = linewidth,
                        alpha = alpha,
                        zorder = zorder)
    
    return artist




def landcover_spatial_moments(name_case: str, sources = [], variables = [], year_start: int = 1995, year_end: int = 2018,
                    moments = ['mean', 'variance']):

    print(f'\nIterate over gridded source data:\n{*sources,}\n')

    from my_resources.sources import query_variables, available_variables
    from my_files.handy import yearly_or_monthly_files, save_df
    from my_files.netcdf import open_netcdf, netcdf_variables_to_array
    from my_resources.units import transform
    from my_gridded.dimensions import select
    from my_math.stats import mean, variance
    from user_in.options_analyses import selected_landcover


    for src in sources:
     
        path                    = query_variables(src, 'path')
        freq_files              = query_variables(src, 'freq_files')

        vars_avail, vars_src    = available_variables(src, variables)

        files                   = yearly_or_monthly_files(freq_files, path, year_start, year_end)

        print('Load data...\n')

        data                    = open_netcdf(files)
        arrays_list             = netcdf_variables_to_array(data, vars_src, 'float32')

        for i_var, var in enumerate(vars_avail):

            array               = arrays_list[i_var]

            array_ux            = transform(array, src, var)

            for lc in selected_landcover:

                print(lc)

                indices_lc      = query_variables(src, 'sel_agg_layer_PFT')[lc]

                array_pft       = select(array_ux, indices_lc,  1)

                array_lc        = mean(array_pft, axes = 0)

                array_mean      = mean(array_lc, axes = (0, -2, -1))

                array_var       = variance(array_lc, axes = (0, -2, -1))

                print(array_mean, array_var)

                








landcover_spatial_moments('cool', sources = ['CLM5-EU3-pft'], variables = ['GPP'])









def query_variables(src, key):

    item                    = variables[src][key]

    return item


def query_grids(src, key):

    item                    = grids[src][key]

    return item


def query_static(src, key):

    item                    = static[src][key]

    return item


def available_variables(src, vars):
    
    variables_src           = variables[src]['var_names']

    vars_avail              = []
    vars_src                = []
    
    for v in vars:
        
        if v not in variables_src.keys(): continue

        var_src             = variables_src[v]
        
        vars_avail.append(v)
        vars_src.append(var_src)

    return vars_avail, vars_src


def query_unit_transform(src, variable, option):

    unit_transform_option   = variables[src][f'var_unit_transform_{option}'][variable]

    return unit_transform_option


from my_.resources.units import sec_per_day


grids = {
    'EU3': {
        'path': '/p/scratch/cjibg31/jibg3105/CESMDataRoot/InputData/share/domains/',
        'file': 'domain.lnd.CLM5EU3_v4.nc',
        'var_lat': 'yc',
        'var_lon': 'xc',
    },
}



static = {
            
            'CLM5-EU3-surf': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/CESMDataRoot/InputData/lnd/clm2/surfdata_map/',
                        'file': 'surfdata_CLM5EU3_v4_pos.nc',
                        'static_vars': ['PCT_NATVEG', 'PCT_CROP', 'PCT_LAKE', 'PCT_WETLAND', 'PCT_NAT_PFT', 'PCT_CFT'],
                        
                        'abbr_PFT': {
                            0: 'Bare Ground',
                            1: 'NET Temperate',
                            2: 'NET Boreal',
                            3: 'NDT Boreal',
                            4: 'BET Tropical',
                            5: 'BET Temperate',
                            6: 'BDT Tropical',
                            7: 'BDT Temperate',
                            8: 'BDT Boreal',
                            9: 'BES Temperate',
                            10: 'BDS Temperate',
                            11: 'BDS Boreal',
                            12: 'C3 arctic grass',
                            13: 'C3 grass',
                            14: 'C4 grass',
                            15: 'UCrop UIrr',
                            16: 'UCrop Irr',
                        },
                        
                        'names_PFT': {
                            0: 'Bare Ground',
                            1: 'Needleleaf evergreen tree - temperate',
                            2: 'Needleleaf evergreen tree - boreal',
                            3: 'Needleleaf deciduous tree - boreal',
                            4: 'Broadleaf evergreen tree - tropical',
                            5: 'Broadleaf evergreen tree - temperate',
                            6: 'Broadleaf deciduous tree - tropical',
                            7: 'Broadleaf deciduous tree - temperate',
                            8: 'Broadleaf deciduous tree - boreal',
                            9: 'Broadleaf evergreen shrub - temperate',
                            10: 'Broadleaf deciduous shrub - temperate',
                            11: 'Broadleaf deciduous shrub - boreal',
                            12: 'C3 arctic grass',
                            13: 'C3 grass',
                            14: 'C4 grass',
                            15: 'C3 Unmanaged Rainfed Crop',
                            16: 'C3 Unmanaged Irrigated Crop',
                        },
                        
                        'sel_agg_layer_PFT': {
                            'ENF': [1, 2],
                            'DBF': [6, 7, 8],
                            'GRA': [12, 13, 14],
                            'CRO': [15, 16],
                        },

                    },

            'hydroclim-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/COSMOREA6/CLIM_CLASS/',
                        'file': 'CLIM_CLASS.nc',
                        'static_vars': ['PRECTmms_COSMOREA6'],
                        'dummy_ax': False,
                    },
}



variables = {
    
            'CLM5-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/CLM5EU3/006/join_8d/',
                        'time_step': '8D',
                        'leap_day': False,
                        'freq_files': 'yearly',
                        'name_label': r'$\mathdefault{CLM5_{grid}}$',
                        
                        'var_names': {
                            'GPP': 'GPP',
                            'NEE': 'NEE',
                            'ER': 'ER',
                            'ET': 'QFLX_EVAP_TOT',
                            'ET-corr': 'QFLX_EVAP_TOT',
                            'LAI': 'TLAI',
                        },

                        'var_units': {
                            'GPP': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'NEE': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ER': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                            'ET-corr': r'$\mathdefault{mm\;day^{-1}}$',
                            'LAI': r'$\mathdefault{m^{2}\;m^{-2}}$',
                        },
                        
                        'var_unit_transform_methods': {
                            'GPP': 'linear',
                            'NEE': 'linear',
                            'ER': 'linear',
                            'ET': 'linear',
                            'ET-corr': 'linear',
                            'LAI': 'linear',

                        },
                        
                        'var_unit_transform_m': {
                            'GPP': sec_per_day,
                            'NEE': sec_per_day,
                            'ER': sec_per_day,
                            'ET': sec_per_day,
                            'ET-corr': sec_per_day,
                            'LAI': 1.0,

                        },

                        'var_unit_transform_b': {
                            'GPP': 0,
                            'NEE': 0,
                            'ER': 0,
                            'ET': 0,
                            'ET-corr': 0,
                            'LAI': 0.0,
                        },

                    },
            
            
            'CLM5-EU3-pft': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/CLM5EU3/006/join_8d/pft/unlim_float/',
                        'time_step':'8D',
                        'leap_day': False,
                        'freq_files':'yearly',
                        'name_label': r'$\mathdefault{CLM5_{PFT}}$',
                        
                        'var_names': {
                            'GPP': 'GPP',
                            'ET': 'ET',
                            'ET-corr': 'ET'
                        },

                        'var_layer': {
                            'GPP': 'PFT',
                            'ET': 'PFT',
                            'ET-corr': 'PFT',
                        },

                        'method_layer_agg': {
                            'PFT': 'mean',
                        },

                        'var_units': {
                            'GPP': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                            'ET-corr': r'$\mathdefault{mm\;day^{-1}}$',
                        },

                        'var_unit_transform_methods': {
                            'GPP': 'linear',
                            'ET': 'linear',
                            'ET-corr': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'GPP': 1,
                            'ET': 1,
                            'ET-corr': 1,
                        },

                        'var_unit_transform_b': {
                            'GPP': 0,
                            'ET': 0,
                            'ET-corr': 0,
                        },

                        'sel_agg_layer_PFT': {
                            'ENF': [1, 2],
                            'DBF': [6, 7, 8],
                            'GRA': [12, 13, 14],
                            'CRO': [15, 16],
                        },
                    },
            
            'GLASS-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/EDWUE/',
                        'time_step': '8D',
                        'leap_day': True,
                        'freq_files':'yearly',
                        'name_label': r'$\mathdefault{GLASS}$',
                        
                        'var_names': {
                            'GPP': 'GPP_GLASS',
                            'ET': 'ET_GLASS',
                            'LAI': 'LAI_GLASS'
                        },

                        'var_units': {
                            'GPP': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                            'LAI': r'$\mathdefault{m^{2}\;m^{-2}}$',
                        },

                        'var_unit_transform_methods': {
                            'GPP': 'linear',
                            'ET': 'linear',
                            'LAI': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'GPP': 1,
                            'ET': 1,
                            'LAI': 1,
                        },

                        'var_unit_transform_b': {
                            'GPP': 0,
                            'ET': 0,
                            'LAI': 0,
                        },
                    },
            
            'ERA5L-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/EDWUE/',
                        'time_step': '8D',
                        'leap_day': True,
                        'freq_files':'yearly',
                        'name_label': r'$\mathdefault{ERA5L}$',
                        
                        'var_names': {
                            'ET': 'ET_ERA5L',
                            'LAI': 'LAI_ERA5L'
                        },

                        'var_units': {
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                            'LAI': r'$\mathdefault{m^{2}\;m^{-2}}$',
                        },

                        'var_unit_transform_methods': {
                            'ET': 'linear',
                            'LAI': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'ET': 1,
                            'LAI': 1,
                        },

                        'var_unit_transform_b': {
                            'ET': 0,
                            'LAI': 0,
                        },
                    },
            
            'GLEAM-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/EDWUE/',
                        'time_step': '8D',
                        'leap_day': True,
                        'freq_files': 'yearly',
                        'name_label': r'$\mathdefault{GLEAM}$',
                        
                        'var_names': {
                            'ET': 'ET_GLEAM',
                        },

                        'var_units': {
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                        },

                        'var_unit_transform_methods': {
                            'ET': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'ET': 1,
                        },

                        'var_unit_transform_b': {
                            'ET': 0,
                        },
                    },
            
            'COSMOREA6-EU3': {
                        'grid': 'EU3',
                        'path': '/p/scratch/cjibg31/jibg3105/data/COSMOREA6/8daily/',
                        'time_step': '8D',
                        'leap_day': True,
                        'freq_files': 'yearly',
                        'name_label': r'$\mathdefault{COSMOREA6}$',

                        'var_names': {
                            'Temp': 'TBOT',
                            'Precip': 'PRECTmms',
                            'SWdown': 'FSDS',
                            'RH': 'RH',
                        },

                        'var_units': {
                            'Temp': r'$\mathdefault{\degree\;C}$',
                            'Precip': r'$\mathdefault{mm\;day^{-1}}$',
                            'RH': r'$\mathdefault{\%}$',
                            'SWdown': r'$\mathdefault{W\;m^{-2}}$'
                        },

                        'var_unit_transform_methods': {
                            'Temp': 'linear',
                            'Precip': 'linear',
                            'RH': 'linear',
                            'SWdown': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'Temp': 1,
                            'Precip': sec_per_day,
                            'RH': 1,
                            'SWdown': 1,
                        },

                        'var_unit_transform_b': {
                            'Temp': -273.15,
                            'Precip': 0,
                            'RH': 0,
                            'SWdown': 0,
                        },
                    },

            'ICOS-WARMWINTER2020': {
                        'processing': 'ONEFLUX',
                        'path': '/p/scratch/cjibg31/jibg3105/data/ICOS/WARMWINTER2020',
                        'leap_day': False,
                        'name_label': r'$\mathdefault{ICOS}$',
                        
                        'date_format': {
                            'H': '%Y%m%d%H%M',
                            'D': '%Y%m%d',
                            'M': '%Y%m',
                        },
                        
                        'var_names': {
                            'GPP': 'GPP_NT_VUT_REF',
                            'NEE': 'NEE_VUT_REF',
                            'ER': 'RECO_NT_VUT_REF',
                            'ET': 'LE_F_MDS',
                            'ET-corr': 'LE_CORR',
                            'Temp': 'TA_F_MDS',
                            'Precip': 'P_F',
                            'RH': 'RH',
                            'SWdown': 'SW_IN_F_MDS',
                        },

                        'var_units': {
                            'GPP': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'NEE': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ER': r'$\mathdefault{g\;C\;day^{-1}}$',
                            'ET': r'$\mathdefault{mm\;day^{-1}}$',
                            'ET-corr': r'$\mathdefault{mm\;day^{-1}}$',
                            'Temp': r'$\mathdefault{\degree\;C}$',
                            'Precip': r'$\mathdefault{mm\;day^{-1}}$',
                            'RH': r'$\mathdefault{\%}$',
                            'SWdown': r'$\mathdefault{W\;m^{-2}}$'
                        },

                        'var_time_step': {
                            'GPP': 'D',
                            'NEE': 'D',
                            'ER': 'D',
                            'ET': 'D',
                            'ET-corr': 'D',
                            'Temp': 'D',
                            'Precip': 'D',
                            'RH': 'H',
                            'SWdown': 'D',
                        },

                        'csv_open_args': {
                            'na_values': ['-9999'], 'index_col': 0
                            },

                        'var_QC': {
                            'GPP': 'NEE_VUT_REF_QC',
                            'NEE': 'NEE_VUT_REF_QC',
                            'ER': 'NEE_VUT_REF_QC',
                            'ET': 'LE_F_MDS_QC',
                            'ET-corr': 'LE_F_MDS_QC',
                            'Temp': 'TA_F_MDS_QC',
                            'Precip': 'P_F_QC',
                            'RH': 'TA_F_MDS_QC',
                            'SWdown': 'SW_IN_F_MDS_QC',
                        },

                        'var_unit_transform_methods': {
                            'GPP': 'linear',
                            'NEE': 'linear',
                            'ER': 'linear',
                            'ET': 'linear',
                            'ET-corr': 'linear',
                            'Temp': 'linear',
                            'Precip': 'linear',
                            'RH': 'linear',
                            'SWdown': 'linear',
                        },
                        
                        'var_unit_transform_m': {
                            'GPP': 1,
                            'NEE': 1,
                            'ER': 1,
                            'ET': 0.035,
                            'ET-corr': 0.035,
                            'Temp': 1,
                            'Precip': 1,
                            'RH': 1,
                            'SWdown': 1,
                        },

                        'var_unit_transform_b': {
                            'GPP': 0,
                            'NEE': 0,
                            'ER': 0,
                            'ET': 0,
                            'ET-corr': 0,
                            'Temp': 0,
                            'Precip': 0,
                            'RH': 0,
                            'SWdown': 0,
                        },
                    },


            'ICOS-ETC-L2-ARCHIVE': {
                        'processing': 'ARCHIVE',
                        'path': '/p/scratch/cjibg31/jibg3105/data/ICOS/ETC_L2_ARCHIVE',
                        'name_label': r'$\mathdefault{ICOS}$',
                        'date_format': '%Y%m%d',
                        
                        'var_names': {
                            'LAI': 'LAI',
                        },

                        'var_units': {
                            'LAI': r'$\mathdefault{m^{2}\;m^{-2}}$',
                        },

                        'csv_open_args': {
                            'na_values': ['-9999'],
                            },

                        'var_unit_transform_methods': {
                            'LAI': 'linear',
                            },
                        
                        'var_unit_transform_m': {
                            'LAI': 1,
                        },

                        'var_unit_transform_b': {
                            'LAI': 0,
                        },
                    },
}



def vertical_top_cax(fx: float = 6.7, fy: float = 13.4, dpi: int = 300, 
                        layout: str = 'constrained', projection = None, frame: bool = False,
                        hspace: float = 0.05, wspace: float = 0.05,
                        annotation: bool = True, x_an: float = -0.05, y_an: float = 1.05): 

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec

    from my_.plot.style import style_1
    from my_.plot.init_ax import init_annotation_ax

    style_1()

    ncols, nrows                    = 2, 5

    figure                          = plt.figure(figsize = (fx, fy), dpi = dpi, layout = layout)

    gs                              = GridSpec(figure = figure, ncols = ncols, nrows = nrows,
                                               height_ratios = [1, 18, 18, 18, 18],
                                               hspace = hspace, wspace = wspace)

    axs_kw                          = {'frameon': frame, 'projection': projection}

    axs                             = gs.subplots(subplot_kw = axs_kw)

    for ax in axs[0,0:2]:
        ax.remove()
    
    axl                             = plt.subplot(gs[0,:], **axs_kw)

    if annotation: init_annotation_ax(axs[1:,:].flatten(), x = x_an, y = y_an)

    return figure, axs[1:,:], axl



import matplotlib as mpl
import matplotlib.pyplot as plt
import cartopy.crs as ccrs


def square(fx: float = 6.7, fy: float = 6.7, dpi: int = 300, projection = None, frame: bool = False):

    import matplotlib.pyplot as plt
    from my_.plot.style import style_1

    style_1()

    fig                             = plt.figure(figsize=(fx, fy), dpi = dpi)

    ax                              = fig.add_subplot(111, projection = projection, frameon = frame)

    return fig, ax


def square_right_cax(fx: float = 6.7, fy: float = 6.7, dpi: int = 300, projection = None, frame: bool = False):

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec
    from my_.plot.style import style_1
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    import matplotlib.axes as maxes
    
    style_1()
    
    nrows = 1
    ncols = 2
    w1, w2 = 10, 1

    fig = plt.figure(figsize=(fx, fy), dpi = dpi, constrained_layout = True)

    gs = GridSpec(figure = fig, 
                    ncols = ncols, nrows = nrows, 
                    width_ratios = [w1, w2],
                    wspace = 0.1)

    ax = fig.add_subplot(gs[0,0], 
                            projection = projection, 
                            frameon = frame)

    divider = make_axes_locatable(ax)
    cax = divider.append_axes('right', 
                                size = '8%', 
                                pad = 0.8, 
                                frameon = frame, 
                                axes_class = maxes.Axes)

    #cax = fig.add_axes([ax.get_position().x1+0.01,ax.get_position().y0,0.02,ax.get_position().height])
    #cax = fig.add_subplot(gs[0,1], frameon = frame)

    return fig, ax, cax


def square_top_cax(fx: float = 6.7, 
                   fy: float = 6.7, 
                   dpi: int = 300, 
                   projection: ccrs.Projection | None = None, 
                   frame: bool = False) -> tuple[plt.Figure, 
                                                 plt.Axes, 
                                                 plt.Axes]:

    from matplotlib.gridspec import GridSpec
    from my_.plot.style import style_1

    style_1()

    
    nrows                           = 2
    ncols                           = 1
    w1, w2                          = 1, 14

    fig                             = plt.figure(figsize=(fx, fy), dpi = dpi)

    gs                              = GridSpec(figure = fig, 
                                                ncols = ncols, nrows = nrows, 
                                                height_ratios = [w1, w2])
    
    ax                              = fig.add_subplot(gs[1, 0], 
                                                      projection = projection, 
                                                      frameon = frame)
    
    cax                             = fig.add_subplot(gs[0,0], frameon = frame)

    return fig, ax, cax, gs


def square_top_right_cax(fx: float = 6.7, 
                         fy: float = 6.7, 
                         dpi: int = 300, 
                         projection: None | ccrs.Projection = None, 
                         frame: bool = False):
    
    from matplotlib.gridspec import GridSpec
    from my_.plot.style import style_1
    
    style_1()

    nrows                           = 2
    ncols                           = 2
    w1, w2                          = 1, 14
    h1, h2                          = 14, 1


    fig                             = plt.figure(figsize=(fx, fy), dpi = dpi)

    gs                              = GridSpec(figure = fig, 
                                                ncols = ncols, nrows = nrows, 
                                                height_ratios = [w1, w2],
                                                width_ratios = [h1, h2])
    
    ax                              = fig.add_subplot(gs[1,0], projection = projection, frameon = frame)
    caxt                            = fig.add_subplot(gs[0,0], frameon = frame)
    caxr                            = fig.add_subplot(gs[1,1], frameon = frame)

    return fig, ax, caxt, caxr



import matplotlib.pyplot as plt
from matplotlib.cm import ScalarMappable

def marker_legend(ax: plt.axes, 
                  dict_labels_markers: dict, 
                  marker_color: str = 'grey', 
                  marker_sizes: float | dict = 10.0, 
                  labelcolor: str = 'k',
                  title: str | None = None,
                  anchor: tuple = (0.5, 0), 
                  markerfirst: bool = True, 
                  fs_labels: float = 12.0, 
                  handletextpad: float = 0.2,
                  columnspacing: float = 0.9, 
                  loc: str = 'lower center', 
                  handlelength: float = 1.2):

    from matplotlib.lines import Line2D
    
    ax.set_yticks([])
    ax.set_xticks([])

    handles = [Line2D([0], [0], 
                       marker = dict_labels_markers[ll], 
                       color = 'white', 
                       markeredgecolor = marker_color, 
                       markersize = marker_sizes[ll], 
                       label = ll) 
                       for ll in dict_labels_markers]
    
    labels = list(dict_labels_markers.keys())

    ax.legend(handles, 
              labels, 
              fontsize = fs_labels, 
              ncol = len(handles),
              frameon = False, 
              title = title,
              labelcolor = labelcolor,
              loc = loc, 
              bbox_to_anchor = anchor, 
              handletextpad = handletextpad, 
              columnspacing = columnspacing,
              handlelength = handlelength, 
              bbox_transform = ax.transAxes, 
              markerfirst = markerfirst)
    

def color_legend(ax: plt.axes,
                 dict_labels_colors: dict,
                 cmap: ScalarMappable | list,
                 linewidth: float = 5.0, 
                 fs_labels: float = 10.0,
                 anchor: tuple = (0.5, 0), 
                 markerfirst: bool = True,
                 handletextpad: float = 0.35,
                 columnspacing: float = 0.9, 
                 loc: str = 'lower center', 
                 handlelength: float = 1.0,
                 ncols: int = 1):

    from matplotlib.lines import Line2D
    
    ax.set_yticks([])
    ax.set_xticks([])

    handles                     = [Line2D([0], [0], color = cmap[v],
                                        linewidth = linewidth, label = k) 
                                        for k, v in dict_labels_colors.items()]

    labels                      = list(dict_labels_colors.keys())

    short_labels                = [l.replace('-EU3', '') for l in labels]

    ax.legend(handles, short_labels, 
            fontsize = fs_labels, 
            ncol = ncols, 
            frameon = False, 
            loc = loc, 
            bbox_to_anchor = anchor, 
            handletextpad = handletextpad, 
            columnspacing = columnspacing,
            handlelength = handlelength, 
            bbox_transform = ax.transAxes, 
            markerfirst = markerfirst)



import pandas as pd

def single_site_model_benchmarks(name, df, variable: str, obs: str, df_static, sel_landcover):

    print('Calculate model benchmarks based on single sites and observations\n')

    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import single_axis_wise
    from my_.math.stats import rmse, pbias, r
    from my_.files.handy import save_df

    df_nona                     = nona_level(df, 
                                             ['Variable', 'Station'],
                                             axis = 1,
                                             how = 'any')

    df_n                        = select_multi_index(df_nona, ['Variable', 'Landcover'],
                                                      keys = [variable, sel_landcover])
    
    df_n.columns = df_n.columns.droplevel(level = [-1, -3])

    df_groups                   = df_n.groupby(axis = 1, level = ['Station'], group_keys = False)

    df_rmse                     = df_groups.apply(single_axis_wise, 
                                                  level = 'Source', 
                                                  key = obs, 
                                                  ffunc = rmse)
    
    df_pbias                    = df_groups.apply(single_axis_wise, level = 'Source', key = obs, ffunc = pbias)
    df_r                        = df_groups.apply(single_axis_wise, level = 'Source', key = obs, ffunc = r)

    names                       = df_static['name']
    lc                          = df_static['landcover']
    vars                        = [variable] * len(lc)

    df_count                    = df_n.count()#.droplevel(['Variable', 'Landcover']).unstack().T.reindex(names)
    
    df_lc_count                 = df_n.groupby(axis = 1, level = ['Source', 'Landcover']).count().sum().T

    new_cols                    = pd.MultiIndex.from_arrays([vars, lc, names])

    df_rmse_sort                = df_rmse.reindex(new_cols, axis = 0)
    df_pbias_sort               = df_pbias.reindex(new_cols, axis = 1).T
    df_r_sort                   = df_r.reindex(new_cols, axis = 1).T

    save_df(df_rmse, f'out/{name}/csv/stations_rmse_{variable}.csv')
    save_df(df_pbias, f'out/{name}/csv/stations_pbias_{variable}.csv')
    save_df(df_r, f'out/{name}/csv/stations_r_{variable}.csv')
    save_df(df_count, f'out/{name}/csv/stations_count_{variable}.csv')
    #save_df(df_lc_count, f'out/{name}/csv/landcover_count_{variable}.csv')
    


def landcover_model_benchmarks(name: str, 
                               df: pd.DataFrame, 
                               variable: str, 
                               obs: str, 
                               sel_landcover: list[str]):

    print('Calculate model benchmarks based on landcover and observations\n')
    
    from my_.series.group import select_multi_index, nona_level
    from my_.series.aggregate import single_level_wise, count_nonzero
    from my_.math.stats import rmse, pbias, r
    from my_.files.handy import save_df

    stations_uq = df.columns.unique(level = 'Station')
    
    stations_count = df.columns\
                       .get_level_values('Station')\
                       .value_counts()
    
    for s in stations_uq:
        
        if stations_count[s] == 1:

            df = df.drop(s, axis = 1, level = 'Station')

    df_nona = nona_level(df, 
                         ['Variable', 'Station'],
                         axis = 1,
                         how = 'any')

    #df_nona = df_nona.dropna(axis = 1, 
    #                         how = 'all')

    df_n = select_multi_index(df_nona, 
                              ['Variable', 'Landcover'],
                              keys = [variable, sel_landcover],
                              axis = 1)
    
    df_groups = df_n.T.groupby(level = ['Landcover'])

    df_groups_rmse = df_groups.apply(single_level_wise, 
                                     level = 'Source', 
                                     key = obs, 
                                     ffunc = rmse,
                                     axis = 0)

    
    df_groups_pbias = df_groups.apply(single_level_wise, 
                                      level = 'Source', 
                                      key = obs, 
                                      ffunc = pbias,
                                      axis = 0)
    
    df_groups_r  = df_groups.apply(single_level_wise, 
                                   level = 'Source', 
                                   key = obs, 
                                   ffunc = r,
                                   axis = 0)

    df_lc_count = df_n.groupby(axis = 1, 
                               level = ['Source', 'Landcover'])\
                                .count()\
                                    .sum()
    
    df_groups_count = pd.DataFrame({'count': 
                                   df_lc_count})

    df_out = df_groups_pbias.join(df_groups_rmse)\
                            .join(df_groups_r)\
                                  .T\
                                  .swaplevel(axis = 1)

    
    save_df(df_groups_count, 
            f'out/{name}/csv/landcover_count_{variable}.csv')
    
    save_df(df_out, 
            f'out/{name}/csv/landcover_r_rmse_pbias_{variable}.csv')


    




import matplotlib as mpl
import matplotlib.pyplot as plt


def save_png(fig: plt.Figure, 
             name_file: str = 'out.png', 
             bbox_inches: str = 'tight', 
             dpi: int = 300, 
             transparent: bool = False):
    
    fig.savefig(name_file, 
                bbox_inches = bbox_inches, 
                dpi = dpi, 
                transparent = transparent)



def save_pdf(pdf, fig):
    
    import matplotlib.pyplot as plt

    pdf.savefig(fig)
    plt.close()


import numpy as np
import os
from my_.files.handy import create_dirs, check_file_exists
from glob import glob
import xarray as xr

forcing_EU3_3h = {
    'name': 'ERA5_forcing_EU3_3h',
    'version': (1, 0, 6),
    'path': '/p/data1/jibg31/FORCINGS/ERA5/',
    'type_file': 'netcdf',
    'year_start': 1950,
    'month_start': 1,
    'year_end': 2022,
    'month_end': 12,
    'resolution_time': '3H',
    'grid': 'EU3',
    'variables': ['P',
                  'Temp',
                  'PSRF',
                  'FSDS',
                  'FLDS',
                  'WIND', 
                  'QBOT',],
    'variable_names': {'P': 'PRECTmms',
                       'Temp': 'TBOT',
                       'PSRF': 'PSRF',
                       'FSDS': 'FSDS',
                       'FLDS': 'FLDS',
                       'WIND': 'WIND',
                       'QBOT': 'QBOT',},
    'variable_dimensions': {'P': ['time', 'lat', 'lon'], 
                            'Temp': ['time', 'lat', 'lon'],
                            'PSRF': ['time', 'lat', 'lon'],
                            'FSDS': ['time', 'lat', 'lon'],
                            'FLDS': ['time', 'lat', 'lon'],
                            'WIND': ['time', 'lat', 'lon'],
                            'QBOT': ['time', 'lat', 'lon'],},
    'variable_units': {'P': 'mm/s',
                       'Temp': 'K',
                       'PSRF': 'Pa',
                       'FSDS': 'W/m^2',
                       'FLDS': 'W/m^2',
                       'WIND': 'm/s',
                       'QBOT': 'kg/kg'},
    'mask_value': None,
    'leapday': False
}



def create_yearly_files(path_rawdata: str,
                        path_out: str,
                        year: int):
    
    create_dirs(path_out)

    files = sorted(glob(f'{path_rawdata}*.nc'))

    if check_file_exists(f'{path_out}/{year}.nc'): print('File is there. Continue...'); return
    
    data_raw = xr.open_mfdataset(files)

    data_raw.to_netcdf(f'{path_out}/{year}.nc',
                       format = 'NETCDF4_CLASSIC', 
                       unlimited_dims = ['time'])




# Notes
# None and none and nones and default values correspond with here?
# Spinup functionality

from dataclasses import dataclass, field

@dataclass
class case:

    """
    I hope to be a generic class for most CLM5 user cases.

    Infos:
    https://www2.cesm.ucar.edu/models/ccsm4.0/clm/models/lnd/clm/doc/UsersGuide/x1327.html
    https://docs.cesm.ucar.edu/models/cesm2/settings/current/drv_input_cesm.html
    https://docs.cesm.ucar.edu/models/cesm2/settings/current/clm5_0_nml.html
    https://docs.cesm.ucar.edu/models/cesm2/settings/current/datm_input.html

    Maybe in future: different datm interpolation methods?
    https://escomp.github.io/ctsm-docs/versions/release-clm5.0/html/users_guide/setting-up-and-running-a-case/customizing-the-datm-namelist.html

    To do:
    - Custom forcings switch: consider the stated stream files in the user datm namelist

    """

    dir_script: str 
    name: str
    compset: str
    n_cpu: int
    year_start: int
    year_end: int
    year_start_forcing: int
    year_end_forcing: int
    dir_domain_lnd: str
    file_domain_lnd: str
    dir_surf: str
    file_surf: str
    months_per_wallclock: int
    time_resolution_forcing_hours: int

    ncpl: int                   = 24
    wallclock: str              = '24:00:00'
    
    mode_datm: str              = 'CLMCRUNCEPv7'
    hist_vars_grid: list        = field(default_factory = list)
    hist_vars_pft: list         = field(default_factory = list)
    hist_frq_grid: int          = -24
    hist_flt_grid: int          = 365
    hist_frq_pft: int           = -24
    hist_flt_pft: int           = 365

    month_start: int            = 1
    month_end: int              = 12

    month_start_forcing: int    = 1
    month_end_forcing: int      = 12

    custom_forcings: bool       = True


    dir_domain_atm: None | str  = None
    file_domain_atm: None | str = None

    dir_init: str | None        = None
    file_init: str | None       = None

    dir_forcing: str | None     = None

    mode_co2: None | str        = None 

    type_co2: str               = 'constant' 
    ppmv_co2: float             = 379.0
    
    series_co2: str             = None

    coldstart: bool             = False

    job_archive: bool           = False

    AD_spin_up: bool            = False

    continue_run: bool          = False


    def create(self):
        
        """
        To create a case directory.

        """
        import os
        import subprocess
        
        if os.path.isdir(f'{self.dir_script}/{self.name}'): 
            
            print('\nCase directory already exists. \n')

            return

        print('\nCreate case...\n')

        subprocess.call([f'{self.dir_script}/create_newcase', 
                         '--case', self.name, 
                         '--res', 'CLM_USRDAT' , 
                         '--compset', self.compset, 
                         '--run-unsupported'], cwd = self.dir_script)
        
        print('\nCase created.\n')
    

    def setup(self, clean: bool = False):
        
        """
        To setup the case and create respective generic namelist files.

        """

        import os
        import subprocess

        dir_setup           = f'{self.dir_script}/{self.name}'

        if clean: subprocess.call([f'{dir_setup}/case.setup', 
                                   '--clean'], cwd = dir_setup)
        
        if os.path.isdir(f'{dir_setup}/CaseDocs/'): 
            
            print('Case is already set-up. Clean?')
            
            return

        print('\nSet-up case ...\n')

        subprocess.call([f'{dir_setup}/xmlchange',
                         f'NTASKS={self.n_cpu}'], cwd = dir_setup)
        
        subprocess.call(f'{dir_setup}/case.setup', cwd = dir_setup)

        print('\nCase set-up.\n')
    
    
    
    def config(self):
     
        """
        Change the config files in the case directory.
        We use the xmlchange functionality.

        """

        print('\nAdjust case configuration...\n')

        import subprocess
        import numpy as np

        dir_setup           = f'{self.dir_script}/{self.name}'

        n_months            = 12 * (self.year_end - self.year_start) - (24 - (self.month_start + self.month_end))

        resubmit            = (np.ceil(n_months / self.months_per_wallclock) - 1).astype(int)

        keys_config         = { 'ATM_DOMAIN_PATH': self.dir_domain_atm,
                                'LND_DOMAIN_PATH': self.dir_domain_lnd,
                                'ATM_DOMAIN_FILE': self.file_domain_atm,
                                'LND_DOMAIN_FILE': self.file_domain_lnd,
                                'DATM_MODE': self.mode_datm,
                                'DATM_CLMNCEP_YR_ALIGN': self.year_start_forcing,
                                'DATM_CLMNCEP_YR_START': self.year_start_forcing,
                                'DATM_CLMNCEP_YR_END': self.year_end_forcing,
                                'RUN_STARTDATE': f'{self.year_start:04d}-{self.month_start:02d}-01',
                                'STOP_OPTION': 'nmonths',
                                'STOP_N': self.months_per_wallclock,
                                'STOP_DATE': -1,
                                'JOB_WALLCLOCK_TIME': self.wallclock,
                                'CLM_ACCELERATED_SPINUP': self.AD_spin_up,
                                'RESUBMIT': resubmit,
                                'CLM_FORCE_COLDSTART': self.coldstart,
                                'ATM_NCPL': self.ncpl,
                                'LND_NCPL': self.ncpl,
                                'DOUT_S': self.job_archive,
                                'CCSM_BGC': self.mode_co2,
                                'CLM_CO2_TYPE': self.type_co2,
                                'CCSM_CO2_PPMV': self.ppmv_co2,
                                'DATM_CO2_TSERIES': self.series_co2,
                                'CONTINUE_RUN': self.continue_run,
                                }
        
        for k, v in keys_config.items():
            
            print(f'Set {k}...')

            if v is None: v = 'none'
            if k in ('CLM_FORCE_COLDSTART', 'CLM_ACCELERATED_SPINUP') and v is False: v = 'off'
            if k in ('CLM_FORCE_COLDSTART', 'CLM_ACCELERATED_SPINUP') and v is True: v = 'on'
            if k in ('DOUT_S', 'CONTINUE_RUN') and v is False: v = 'FALSE'
            if k in ('DOUT_S', 'CONTINUE_RUN') and v is True: v = 'TRUE'

            subprocess.call([f'{dir_setup}/xmlchange',
                         f'{k}={v}'], cwd = dir_setup)
            
        print('\nCase configuration adjusted.\n')


    def namelists(self):

        """
        Change the user namelist files in the case directory.
        We use blueprint files from this repository, change them,
        and copy them to the case directory.

        """

        print('\nCreate case namelists...\n')

        dir_setup           = f'{self.dir_script}/{self.name}'
        
        file_init           = f"finidat = '{self.dir_init}/{self.file_init}'" if self.file_init is not None else ''

        hist_vars_grid      = "'" + "', '".join(self.hist_vars_grid) + "'"

        dt_limit            = (24 + self.timestep_forcing_hours) / self.timestep_forcing_hours

        if self.hist_vars_pft:

            hist_vars_pft_0 = f'hist_dov2xy(2) = .false.'
            hist_vars_pft_1 = f"hist_type1d_pertape(2) = 'PFTS'"
            hist_vars_pft_2 = f'hist_nhtfrq(2) = {self.hist_frq_pft}'
            hist_vars_pft_3 = f'nhist_mfilt(2) = {self.hist_flt_pft}'
            hist_vars_pft_4 = "'" + "', '".join(self.hist_vars_pft) + "'"

            hist_pft_str    = '\n'.join([hist_vars_pft_0, 
                                        hist_vars_pft_1,
                                        hist_vars_pft_2,
                                        hist_vars_pft_3,
                                        hist_vars_pft_4,])
        else:
            
            hist_pft_str    = ''


        keys_namelist_clm   = { 'file_surf': f"'{self.dir_surf}/{self.file_surf}'",
                                'file_init': file_init,
                                'hist_frq': self.hist_frq_grid,
                                'hist_flt': self.hist_flt_grid,
                                'hist_vars_grid': hist_vars_grid,
                                'hist_vars_pft': hist_pft_str,
                                }
        
        keys_namelist_datm  = { 'year_start': self.year_start_forcing,
                                'year_end': self.year_end_forcing,
                                'dt_limit': dt_limit,
                                }

        self.search_replace('config_files/user_nl/', 'user_nl_clm', keys_namelist_clm, dir_setup)
        self.search_replace('config_files/user_nl/', 'user_nl_datm', keys_namelist_datm, dir_setup)


        print('\nNamelists created and copied.\n')
    

    def search_replace(self, path_in, name_file, search_replace_dict, path_out):
        
        with open(f'{path_in}/{name_file}', 'r') as file:
        
            content         = file.read()

            for k, v in search_replace_dict.items():

                content     = content.replace(k, str(v))

        with open(f'{path_out}/{name_file}', 'w') as file:

            file.write(content)


    def streams(self):

        """
        Create and adjust the datm stream files files.
        This is to run CLM5 with your own forcings.
        We use the blueprint files in this repository, change them,
        and copy them to the case directory.

        """

        if not self.custom_forcings: print('\nNo stream files necessary.\n'); return
        
        print('\nCreate case stream files...\n')

        import subprocess

        dir_setup           = f'{self.dir_script}/{self.name}'

        year_files          = [f'{y}-{m:02d}.nc' for y in range(self.year_start_forcing, self.year_end_forcing + 1) for m in range(1,13)]

        keys_streams        = { 'dir_domain': self.dir_domain_atm,
                                'file_domain': self.file_domain_atm, 
                                'dir_forcing': self.dir_forcing,
                                'year_files': '\n'.join(year_files),
                               }
        
        self.search_replace('config_files/user_datm/', 'user_datm.streams.txt.CLMCRUNCEPv7.Precip', keys_streams, dir_setup)
        self.search_replace('config_files/user_datm/', 'user_datm.streams.txt.CLMCRUNCEPv7.Solar', keys_streams, dir_setup)
        self.search_replace('config_files/user_datm/', 'user_datm.streams.txt.CLMCRUNCEPv7.TPQW', keys_streams, dir_setup)

        subprocess.call(f'{dir_setup}/preview_namelists', cwd = dir_setup)
        subprocess.call(f'cp user_datm.streams.* Buildconf/datmconf/', shell = True, cwd = dir_setup)

        print('\nCase stream files created.\n')


    def build(self, clean: bool = False):

        """
        Build the model case with the case.build script.

        """

        import subprocess
        import mmap

        dir_setup           = f'{self.dir_script}/{self.name}'

        with open(f'{dir_setup}/CaseStatus', 'r') as file, \
            mmap.mmap(file.fileno(), 0, access = mmap.ACCESS_READ) as s:
            
            if s.find(b'case.build success') != -1:

                print('Case was already built. Clean?')
                
                return

        if clean:

            print('Cleaning case build...')

            subprocess.call([f'{dir_setup}/case.build', '--clean-all'], cwd = dir_setup)

            print('Case build cleaned.')

        print('Build case...')

        subprocess.call([f'{dir_setup}/case.build'], cwd = dir_setup)

        print('Case built.')


    def submit(self):

        """
        Submit your case for simulation with the case.submit script.
        
        """

        import subprocess

        dir_setup           = f'{self.dir_script}/{self.name}'

        print('\nCase submit...\n')

        subprocess.call([f'{dir_setup}/case.submit'], cwd = dir_setup)

        print('\nCase submitted. All done!\n')


@dataclass
class restart:

    """
    Yes, under heavy contrsuction, does not work yet

    """

    
    dir_output: str
    files_output : list | str
    n_lat: int
    n_lon: int

    names_pft: list = field(default_factory=lambda: ['BG',      #1
                                                    'NEF_Te',   #2
                                                    'NEF_Bo',   #3
                                                    'NDF_Bo',   #4
                                                    'BEF_Tr',   #5
                                                    'BEF_Te',   #6
                                                    'BDF_Tr',   #7
                                                    'BDF_Te',   #8
                                                    'BDT_Bo',   #9
                                                    'BES_Te',   #10
                                                    'BDS_Te',   #11
                                                    'BDS_Bo',   #12
                                                    'C3arct',   #13
                                                    'C3',       #14
                                                    'C4',       #15
                                                    'C3_rain',  #16
                                                    'C3_irr'])  #17
    


    def dataset(self):

        import xarray as xr
        from glob import glob

        files   = glob(f'{self.dir_output}/{self.files_output}')

        args    = {'chunks': {'time': 365}, 'decode_times': True}

        dataset = xr.open_dataset(files[0], **args) if len(files) == 1 else xr.open_mfdataset(files, **args)

        return dataset


    def to_sparse(self, data, vegtype, jxy, ixy, shape):

        import numpy as np
        import sparse as sp

        """
        Takes an input numpy array and converts it to a sparse array.

        Parameters
        ----------
        data: numpy.ndarray
            1D or 2D Data stored in compressed form.
        vegtype: numpy.ndarray

        jxy: numpy.ndarray
            Latitude index
        ixy: numpy.ndarray
            Longitude index
        shape: tuple
            Shape provided as sizes of (vegtype, jxy, ixy) in uncompressed
            form.

        Returns
        -------
        sparse.COO
            Sparse nD array
        """

        # This constructs a list of coordinate locations at which data exists
        # it works for arbitrary number of dimensions but assumes that the last dimension
        # is the "stacked" dimension i.e. "pft"

        if ((data.ndim == 1)):

            coords = np.stack([vegtype, jxy - 1, ixy - 1], axis = 0)
           
        else:

            raise NotImplementedError

        sparse_coo          = sp.COO(coords = coords,
                                    data = data.ravel(),
                                    shape = data.shape[:-1] + shape,
                                    fill_value = np.nan)

        return sparse_coo


    def convert_pft_variables_to_sparse(self):

        import xarray as xr
        import numpy as np
        import sparse as sp

        from my_resources.CLM5_carbon_pools import pft, soil

        """
        Convert 2D PFT variables in dataset to 4D sparse arrays.

        Parameters
        ----------
        dataset: xarray.Dataset
            Dataset with DataArrays that have a `pft` dimension.

        Returns
        -------
        xarray.Dataset
            Dataset whose "PFT" arrays are now sparse arrays
            with `pft` dimension expanded out to (type, lat, lon)
        """

        variables = pft #| soil

        dataset = self.dataset()

        # extract PFT variables
        pfts = xr.Dataset({k: v for k, v in dataset.items() if (("pft" in v.dims) and (k in variables.keys()))})

        # extract coordinate index locations
        ixy = dataset.pfts1d_ixy.astype(int)
        jxy = dataset.pfts1d_jxy.astype(int)
        vegtype = dataset.pfts1d_itypveg.astype(int)
        npft = len(self.names_pft)

        # expected shape of sparse arrays to pass to `to_sparse` (excludes time)
        output_sizes = {'vegtype': npft, 'lat': self.n_lat, 'lon': self.n_lon}

        result = xr.Dataset()
        # we loop over variables so we can specify the appropriate dtype
        for var in pfts:
            result[var] = xr.apply_ufunc(
                self.to_sparse,
                pfts[var],
                vegtype,
                jxy,
                ixy,
                kwargs = {'shape': tuple(output_sizes.values())},
                input_core_dims = [["pft"]] * 4,
                output_core_dims = [["vegtype", "lat", "lon"]],
                dask = "parallelized",
                dask_gufunc_kwargs = dict(
                    meta = sp.COO(np.array([], dtype = pfts[var].dtype)),
                    output_sizes = output_sizes,
                ),
                keep_attrs = True,
            )

        # copy over coordinate variables lat, lon
        # result = result.update(dataset[["lat", "lon"]])

        result["vegtype"] = self.names_pft

        # save the dataset attributes
        result.attrs = dataset.attrs

        return result






def src_var_apply_nc(name: str = '', sources: list = [], variables: list = [], 
                    array_sel_4d = None, method_sel = 'argmax', axis_sel: int = 0, 
                    axis_sel_shrink: int = 0, method_agg_4d: str = 'mean',
                    method_agg_time: str ='mean'):

    """
    Make an array 4d tesseract ("cube")
    assuming axis -2 and -1 are lat lon
    aggregated by method keywords 
    over time and e.g. pfts or soil layers
    """

    import xarray as xr

    from my_files.netcdf import open_netcdf
    from my_files.handy import check_file_exists
    from my_resources.sources import query_variables
    from my_resources.units import transform
    from my_files.netcdf import netcdf_variable_to_array
    from my_gridded.aggregate import select_o_apply_d


    file_out                    = f"out/nc/{name}_*{'+'.join(sources)}*_{'+'.join(variables)}_{method_agg_time}.nc"

    print(file_out)

    if check_file_exists(file_out): return

    print('File does not exists')

    data_out                    = xr.Dataset({})
    
    for src in sources:

        path                    = query_variables(src = src, key = 'path')
        data_src                = open_netcdf(f'{path}*.nc')

        for var in variables:

            print(f'Processing {src} - {var}')
            
            var_src                     = query_variables(src = src,  key = 'var_names')[var]
            var_array                   = netcdf_variable_to_array(data_src, var_src, dtype = 'float32')

            var_array_ux                = transform(var_array, src, var)
            
            var_array_2d                = select_o_apply_d(var_array_ux, array_sel_4d = array_sel_4d,
                                                           method_sel = method_sel, axis_sel = axis_sel,
                                                           axis_sel_shrink = axis_sel_shrink, 
                                                           method_agg_4d = method_agg_4d, 
                                                           method_agg_time = method_agg_time)
            
            data_out[f'{src}_{var}_{method_agg_time}'] = (['latitude', 'longitude'], var_array_2d)

    print(f'Succesfully created tesseract for sources and variables!\n')

    data_out.to_netcdf(f'{file_out}', mode='w', format='NETCDF4_CLASSIC')


import numpy as np

def repeat_nd(array: np.ndarray,
              dims: int | list,
              repeats: dict,
              ndims: int =  2,) -> np.ndarray:
    
    if isinstance(dims, int): dims = [dims]

    ddims = [d for d in range(ndims) 
             if d not in dims]

    slx = [slice(0, None)] * ndims
    
    for d in ddims: slx[d] = None

    array = array[tuple(slx)]

    for d in ddims: 
        
        array = np.repeat(array, 
                          repeats[d], 
                          d)

    return array



import pandas as pd
import numpy as np

def peaks(data: pd.Series | np.ndarray, 
          sigma: int = 6,
          distance: int = 46):

    from scipy.signal import find_peaks
    from scipy.ndimage import gaussian_filter1d

    smoo_ = gaussian_filter1d(data, sigma = sigma)

    peaks_ = find_peaks(smoo_, distance = distance)

    if isinstance(peaks_[0], np.ndarray):
        return peaks_[0]
    elif not peaks_[0]: 
        return np.nan
    elif isinstance(data, pd.Series):
        return peaks_[0][0]
    elif isinstance(data, np.ndarray):
        return peaks_[0]


def inflictions(df: pd.Series,
                sigma: int = 6):
    
    from scipy.ndimage import gaussian_filter1d

    smoo = gaussian_filter1d(df, sigma = sigma)

    d2 = np.gradient(np.gradient(smoo))

    infl = np.nonzero(np.diff(np.sign(d2)))[0]

    df_out = df.copy().drop(df.index)

    if (len(infl) == 0) or (len(infl) >= len(df.index) / 2): 
        
        infl = [np.nan]

    for i in range(len(infl)):

        df_out[i] = infl[i]

    return df_out




import numpy as np
import matplotlib.pyplot as plt
import cartopy.crs as ccrs



def EU3_plot_lines(ax, pc, xs, ys, 
                   lw_grid: float = 0.8, 
                   lw_coast: float = 0.8, 
                   color_grid: str = 'grey', 
                   ls_grid: str = '--', 
                   xticks = np.arange(-40, 80, 20), 
                   yticks = np.arange(-20, 80, 20),
                   fs_label: float = 8, 
                   label_sides: list = ['bottom', 'right']):
    
    import matplotlib.ticker as mticker

    ax.coastlines(linewidth = lw_coast, zorder=2)
    
    gl = ax.gridlines(crs = pc, 
                      linewidth = lw_grid,
                      color = color_grid, 
                      linestyle = ls_grid, 
                      draw_labels = True, 
                      x_inline = False, 
                      y_inline = False, 
                      zorder = 3)
    
    if 'right' not in label_sides: 
        gl.right_labels         = False
    if 'bottom' not in label_sides:
        gl.bottom_labels        = False
    if 'top' not in label_sides:
        gl.top_labels           = False
    if 'left' not in label_sides:
        gl.left_labels          = False
    
    gl.xlocator = mticker.FixedLocator(xticks)
    gl.ylocator = mticker.FixedLocator(yticks)
    gl.xlabel_style = {'size': fs_label}
    gl.ylabel_style = {'size': fs_label}

    ax.set_extent([*xs,*ys])


def map_point_locations(ax: plt.Axes, 
                        lats: np.ndarray, 
                        lons: np.ndarray,  
                        size_marker: list = [], 
                        marker: str = 'x', 
                        color: str | list = [],
                        edgecolor: str | list = [], 
                        projection: None | ccrs.Projection = None, 
                        zorder: int = 5, 
                        alpha: float = 0.8,
                        alpha_land: float = 1.0, 
                        alpha_ocean = 0.3, 
                        land_color: str | None = 'dimgrey', 
                        ocean_color: str | None = None):

    import cartopy.feature as cfeature
    from my_.plot.basic import scatter

    ax.add_feature(cfeature.LAND, 
                   color = land_color, 
                   alpha = alpha_land, 
                   zorder = 0)
    
    ax.add_feature(cfeature.OCEAN, 
                   color= ocean_color, 
                   alpha = alpha_ocean, 
                   zorder = 0)
    
    for i, (lat, lon) in enumerate(zip(lats, lons)):
    
        ax.plot(lon, lat,
                marker = marker,
                markersize = size_marker[0],
                color = color,
                markeredgecolor = edgecolor[i],
                transform = projection,
                alpha = alpha,
                zorder = zorder)

    #artist = scatter(ax, 
    #                 lons, 
    #                 lats, 
    #                 sizes_marker = size_marker, 
    #                 marker = marker, 
    #                 colors_marker = color,
    #                 edgecolor = edgecolor,
    #                 projection = projection, 
    #                 alpha = alpha, 
    #                 zorder = zorder)

    #return artist



def count_valid(array):

    import numpy as np

    valid_values                = ~np.isnan(array)

    count_valid                 = np.count_nonzero(valid_values)

    return count_valid


def count_value_occ(array, value):

    import numpy as np

    valid_values                = np.where(array == value, True, False)

    count_valid                 = np.count_nonzero(valid_values)

    return count_valid



import pandas as pd
import xarray as xr
    
def resample(df: pd.DataFrame | pd.Series | xr.Dataset | xr.DataArray,
            offset_str: str, 
            method: str = 'mean',
            yearly_agg: bool = True,
            **kwargs) -> pd.DataFrame | pd.Series | xr.Dataset | xr.DataArray:
    
    print(f'\nResampling dataframe by {method} to {offset_str} time frequency...\n')

    if isinstance(df, xr.Dataset) | isinstance(df, xr.DataArray):

        func_ = getattr(xr.DataArray, method)
    
        index = 'time.year'
        
        offset_str_ = {'time': offset_str}
        
        df_resampler = df.resample(indexer = offset_str_)

        df_resampled = func_(df_resampler, **kwargs)

    elif isinstance(df, pd.DataFrame) | isinstance(df, pd.Series):
    
        index = df.index.year

        if yearly_agg: df = df.groupby(index, group_keys = False)

        df_resampler = df.resample(offset_str)

        df_resampled = df_resampler.agg(method)

    else: NotImplementedError('')

    return df_resampled


def reindex(df, timeseries, method =  None):

    df_new_time = df.reindex(timeseries, method = method)

    return df_new_time

    


import pandas as pd
import os

def check_file_exists(str_file):

    from glob import glob

    if len(glob(str_file)) > 0:
        
        print(f'\nOutput file(s) {str_file} exists. Return.\n')
        
        return True
    
    else:
        
        print(f'\nOutput file(s) {str_file} does not exist.')
        print('It will be created now...\n')
        
        return False


def yearly_files(path, y0: int = 1995, y1: int = 2018):

    files                   = [f"{path}/{str(y)}.nc" for y in range(y0, y1+1)]

    return files


def monthly_files(path, y0: int = 1995, y1: int = 2018):

    files                   = [f"{path}/{str(y)}-{str(m).zfill(2)}.nc" for y in range(y0, y1+1) for m in range(1,13)]

    return files


def yearly_or_monthly_files(freq_files, path, y0, y1):

    if freq_files == 'monthly':

        files               = monthly_files(path, y0, y1)

    elif freq_files == 'yearly':

        files               = yearly_files(path, y0, y1)

    return files


def save_df(df: pd.DataFrame, 
            name: str, 
            format: str = 'csv', 
            mode: str = 'w'):

    if format == 'csv':
            
        df.to_csv(name)
        
    elif format == 'parquet':
           
        df.to_parquet(name)


def open_csv_or_parquet(file: str, csv_args = {'index_col': 0}, parquet_args = {}):

    from my_.files.csv import open_csv
    from my_.files.parquet import open_parquet

    ending                      = file.split('.')[-1]
    
    open_method                 = open_csv if ending == 'csv' else open_parquet

    args                        = csv_args if ending == 'csv' else parquet_args

    df                          = open_method(file, **args)

    return df


def read_resample_save(case_name: str, files: list, origin: str, suffixes: list = [], file_format: str = 'parquet',
                        join: str = 'outer', offset_str: str = 'D', interp_method: str = 'mean'):

    from glob import glob
    from my_.series.interpolate import resample
    from my_.series.aggregate import concat
    from my_.resources.sources import variables
    from my_.series.group import layer_level_to_int

    
    file_out                    = f'out/{case_name}/{file_format}/Resampled_{origin}.{file_format}'

    list_dfs                    = []

    if glob(file_out): print('Resampled file found\n'); return open_csv_or_parquet(file_out)

    print('Resampling...\n')

    for i_f, f in enumerate(files):

        if len(suffixes) > 0:

            suffix              = f'_{suffixes[i_f]}'
        
        else: suffix = ''

        df                  = open_csv_or_parquet(f).add_suffix(suffix)

        src                 = df.columns.get_level_values('Source').unique()[0]

        if 'var_layer' in variables[src]:

            df_int          = layer_level_to_int(df)

            df_groups       = df_int.groupby(axis = 1, level = ['Source', 'Variable', 'Landcover', 'Station'])

            df              = df_groups.apply(aggregate_layer)

        list_dfs.append(df)
    
    df                      = concat(list_dfs, join = join, axis = 1)

    #print(df.loc[((df.index > pd.to_datetime('1995-01-01')) & (df.index < pd.to_datetime('2018-12-31')))].groupby(axis = 1, level = ['Source', 'Landcover']).count().sum())

    df_resampled            = resample(df, offset_str, method = interp_method)
    #print(df_resampled[((df_resampled.index > pd.to_datetime('1995-01-01')) & (df_resampled.index < pd.to_datetime('2018-12-31')))].groupby(axis = 1, level = ['Source', 'Landcover']).count().sum())

    save_df(df_resampled, file_out, file_format)

    return df_resampled


def aggregate_layer(df):

    from my_.resources.sources import query_variables

    from my_.series.group import select_multi_index

    from user_in.options_analyses import selected_landcover


    src                     = df.columns.get_level_values('Source').unique()[0]

    var                     = df.columns.get_level_values('Variable').unique()[0]

    lc                      = df.columns.get_level_values('Landcover').unique()[0]

    if lc not in selected_landcover: return df.agg('mean', axis = 1)

    type_layer              = query_variables(src, 'var_layer')[var]

    method_agg              = query_variables(src, 'method_layer_agg')[type_layer]

    if type_layer == 'PFT':

        selection           = query_variables(src, f'sel_agg_layer_PFT')[lc]

        df_sel              = select_multi_index(df, 'Layer', [selection])

        df_agg              = df_sel.agg(method_agg, axis = 1)

    return df_agg

    
def create_dirs(dirs: str):

    from pathlib import Path

    if isinstance(dirs, str): 
        
        dirs_= [dirs]

    else: 

        dirs_ = dirs
        
    for dir in dirs_:

        Path(dir).mkdir(parents = True, exist_ok = True)





import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from my_.plot.init_ax import init_annotation_ax
from my_.plot.style import style_1

def horizontal(fx: float = 6.7, fy: float = 3.35, dpi: int = 300, projection = None, frame: bool = False):

    import matplotlib.pyplot as plt
    from my_.plot.style import style_1

    style_1()

    nrows                           = 1
    ncols                           = 2

    fig                             = plt.figure(figsize=(fx,fy), dpi= dpi)

    ax1                             = fig.add_subplot(nrows, ncols, 1, projection = projection, frameon = frame)
    ax2                             = fig.add_subplot(nrows, ncols, 2, projection = projection, frameon = frame)

    return fig, [ax1, ax2]


def horizontal_right_cax(fx: float = 6.7, fy: float =3.35, dpi: int = 300, projection = None, frame: bool = False):

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec
    from my_.plot.style import style_1

    style_1()

    nrows                           = 1
    ncols                           = 3
    w1, w2                          = 14, 1

    fig                             = plt.figure(figsize=(fx,fy), dpi= dpi)

    gs                              = GridSpec(figure = fig, 
                                                ncols = ncols, nrows = nrows, 
                                                width_ratios = [w1, w1, w2], wspace = 0.3)

    ax1                             = fig.add_subplot(gs[0, 0], projection = projection, frameon = frame)
    ax2                             = fig.add_subplot(gs[0, 1], projection = projection, frameon = frame)

    cax                             = fig.add_subplot(gs[0, 2], frameon = frame)

    return fig, [ax1, ax2], cax


def horizontal_top_cax(fx: float = 6.7, fy: float = 3.4, dpi: int = 300, projection = None,
                       frame: bool = False, annotation: bool = True, hspace = 0.1,
                       x_an: float = 0.05, y_an: float = 1.05):

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec
    from my_.plot.init_ax import init_annotation_ax
    from my_.plot.style import style_1

    style_1()

    nrows                           = 2
    ncols                           = 2
    w1, w2                          = 1, 14

    fig                             = plt.figure(figsize=(fx,fy), dpi= dpi)

    gs                              = GridSpec(figure = fig, 
                                                ncols = ncols, nrows = nrows, 
                                                height_ratios = [w1, w2],
                                                hspace = hspace)

    ax1                             = fig.add_subplot(gs[1, 0], projection = projection, frameon = frame)
    ax2                             = fig.add_subplot(gs[1, 1], projection = projection, frameon = frame)

    cax                             = fig.add_subplot(gs[0, :], frameon = frame)

    if annotation: init_annotation_ax([ax1, ax2], x = x_an, y = y_an)

    return fig, [ax1, ax2], cax


def horizontal_2top_cax(fx: float = 6.7, 
                        fy: float = 3.4, 
                        dpi: int = 300, 
                        projection = None,
                        frame: bool = False, 
                        annotation: bool = False, 
                        hspace = 0.1,
                        x_an: float = 0.05, 
                        y_an: float = 1.05):


    style_1()

    nrows = 2
    ncols = 2
    w1, w2 = 1, 14

    fig = plt.figure(figsize = (fx, fy), 
                     dpi = dpi)

    gs = GridSpec(figure = fig, 
                  ncols = ncols, 
                  nrows = nrows, 
                  height_ratios = [w1, w2],
                  hspace = hspace)

    ax1 = fig.add_subplot(gs[1, 0], 
                          projection = projection, 
                          frameon = frame)
    
    ax2 = fig.add_subplot(gs[1, 1], 
                          projection = projection, 
                          frameon = frame)

    cax1 = fig.add_subplot(gs[0, 0], frameon = frame)
    cax2 = fig.add_subplot(gs[0, 1], frameon = frame)

    if annotation: init_annotation_ax([ax1, ax2], x = x_an, y = y_an)

    return fig, [ax1, ax2], [cax1, cax2]

def horizontal_topleft_cax(fx: float = 6.7, 
                           fy: float = 3.4, 
                           dpi: int = 300, 
                           projection = None,
                           frame: bool = False, 
                           annotation: bool = False, 
                           hspace = 0.1,
                           x_an: float = 0.05, 
                           y_an: float = 1.05):


    style_1()

    nrows = 2
    ncols = 2
    w1, w2 = 1, 14

    fig = plt.figure(figsize = (fx, fy), 
                     dpi = dpi)

    gs = GridSpec(figure = fig, 
                  ncols = ncols, 
                  nrows = nrows, 
                  height_ratios = [w1, w2],
                  hspace = hspace)

    ax1 = fig.add_subplot(gs[1, 0], 
                          projection = projection, 
                          frameon = frame)
    
    ax2 = fig.add_subplot(gs[1, 1], 
                          projection = projection, 
                          frameon = frame)

    cax1 = fig.add_subplot(gs[0, 0], frameon = frame)

    if annotation: init_annotation_ax([ax1, ax2], x = x_an, y = y_an)

    return fig, [ax1, ax2], cax1



import os
import pandas as pd
from dataclasses import dataclass
from my_.data.sites import from_network, information

WARMWINTER_2020_daily = {
    'name': 'WARMWINTER_2020_daily',
    'version': (1, 0, 6),
    'path': '/p/scratch/cjibg31/jibg3105/data/ICOS/WARMWINTER2020/',
    'type_file': 'csv',
    'time_request': 'D',
    'year_start': 1996,
    'month_start': 1,
    'year_end': 2020,
    'month_end': 12,
    'date_formats': {'H': '%Y%m%d%H%M',
                     'D': '%Y%m%d',
                     'M': '%Y%m',},
    'variables': ['GPP',
                  'NEE',
                  'ER',
                  'LE',
                  'LE-corr',
                  'Temp',
                  'P',
                  'RH',
                  'FSDS',
                  'NetRad',
                  'SoiH',
                  'VPD'],
    'variable_names': {'GPP': 'GPP_NT_VUT_REF',
                       'NEE': 'NEE_VUT_REF',
                       'ER': 'RECO_NT_VUT_REF',
                       'LE': 'LE_F_MDS',
                       'LE-corr': 'LE_CORR',
                       'Temp': 'TA_F_MDS',
                       'Precip': 'P_F',
                       'RH': 'RH',
                       'SWdown': 'SW_IN_F_MDS',
                       'NetRad': 'NETRAD',
                       'SoiH': 'G_F_MDS',
                       'VPD': 'VPD_F_MDS'},
    'variable_QC': {'GPP': 'NEE_VUT_REF_QC',
                    'NEE': 'NEE_VUT_REF_QC',
                    'ER': 'NEE_VUT_REF_QC',
                    'LE': 'LE_F_MDS_QC',
                    'LE-corr': 'LE_F_MDS_QC',
                    'Temp': 'TA_F_MDS_QC',
                    'P': 'P_F_QC',
                    'RH': 'TA_F_MDS_QC',
                    'FSDS': 'SW_IN_F_MDS_QC',
                    'NetRad': 'NETRAD_QC',
                    'SoiH': 'G_F_MDS_QC',
                    'VPD': 'VPD_F_MDS_QC'},
    'variable_dimensions': {'GPP': ['time'],
                            'NEE': ['time'],
                            'ER': ['time'],
                            'LE': ['time'],
                            'LE-corr': ['time'],
                            'Temp': ['time'],
                            'P': ['time'],
                            'RH': ['time'],
                            'FSDS': ['time'],
                            'NetRad': ['time'],
                            'SoiH': ['time'],
                            'VPD': ['time'],},
    'variable_units': {'GPP': {'H': 'umolCO2 m^-2 s^-1',
                               'D': 'g C / (m^2 day)',
                               '7D': 'g C / (m^2 day)',
                               'M': 'g C / (m^2 day)',
                               'Y': 'g C / (m^2 year)'},
                       'NEE': {'H': 'umolCO2 m^-2 s^-1',
                               'D': 'g C / (m^2 day)',
                               '7D': 'g C / (m^2 day)',
                               'M': 'g C / (m^2 day)',
                               'Y': 'g C / (m^2 year)'},
                       'ER': {'H': 'umolCO2 m^-2 s^-1',
                              'D': 'g C / (m^2 day)',
                              'M': 'g C / (m^2 day)',
                              '7D': 'g C / (m^2 day)',
                              'Y': 'g C / (m^2 year)'},
                       'LE': {'H': 'W / m^2',
                              'D': 'W / m^2',
                              '7D': 'W / m^2',
                              'M': 'W / m^2',
                              'Y': 'W / m^2'},
                       'LE-corr': {'H': 'W / m^2',
                                   'D': 'W / m^2',
                                   '7D': 'W / m^2',
                                   'M': 'W / m^2',
                                   'Y': 'W / m^2'},
                       'Temp': {'H': 'degree C',
                                'D': 'degree C',
                                '7D': 'degree C',
                                'M': 'degree C',
                                'Y': 'degree C'},
                       'P': {'H': 'mm',
                             'D': 'mm / day',
                             '7D': 'mm / day',
                             'M': 'mm / day',
                             'Y': 'mm / year'},
                       'RH': {'H': '%'},
                       'FSDS': {'H': 'W / m^2',
                                'D': 'W / m^2',
                                '7D': 'W / m^2',
                                'M': 'W / m^2',
                                'Y': 'W / m^2'},
                       'NetRad': {'H': 'W / m^2',
                                  'D': 'W / m^2',
                                  '7D': 'W / m^2',
                                  'M': 'W / m^2',
                                  'Y': 'W / m^2'},
                       'SoiH': {'H': 'W / m^2',
                                'D': 'W / m^2',
                                '7D': 'W / m^2',
                                'M': 'W / m^2',
                                'Y': 'W / m^2'},
                       'VPD': {'H': 'hPa',
                               'D': 'hPa',
                               '7D': 'hPa',
                               'M': 'hPa',
                               'Y': 'hPa'}},
    'variable_resolutions': {'GPP': ['H', 'D', '7D', 'M', 'Y'],
                             'NEE': ['H', 'D', '7D', 'M', 'Y'],
                             'ER': ['H', 'D', '7D', 'M', 'Y'],
                             'ET': ['H', 'D', '7D', 'M', 'Y'],
                             'ET-corr': ['H', 'D', '7D', 'M', 'Y'],
                             'Temp': ['H', 'D', '7D', 'M', 'Y'],
                             'P': ['H', 'D', '7D', 'M', 'Y'],
                             'RH': ['H'],
                             'FSDS': ['H', 'D', '7D', 'M', 'Y'],
                             'NetRad': ['H', 'D', '7D', 'M', 'Y'],
                             'SoiH': ['H', 'D', '7D', 'M', 'Y'],
                             'VPD': ['H', 'D', '7D', 'M', 'Y'],},
    'mask_value': -9999,
    'leapday': False,}

@dataclass
class ONEFLUX:
    
    name: str
    version: tuple[int, int, int]
    path: os.PathLike
    type_file: str
    year_start: int
    time_request: str
    month_start: int
    year_end: int
    month_end: int
    date_formats: dict
    variables: dict
    variable_names: dict
    variable_QC: dict
    variable_dimensions: dict
    variable_units: dict
    variable_resolutions: dict
    mask_value: list | int | float | str | None
    leapday: bool

    def sites_info(self):

        sites = from_network(network = 'ICOS',
                             verified = False)
        
        names_short = [information(s)['attributes']['general']['shortName']
                      for s in sites]
        
        self.ids = sites
        self.names_short = names_short


    def open_data(self):

        order = ['H', 'D', '7D', 'M', 'Y']

        request_time_available = [k for k, v in self.variable_resolutions.items()
                                  if self.time_request in v]
        
        request_time_missing = [k for k in self.variables
                                if k not in request_time_available]

        
        


        
        #data_ = pd.read_csv()


    def quality_filter(self):
        
        ...


    def le_2_et(self):
        ...

    

if __name__ == '__main__':

    data_ = ONEFLUX(**WARMWINTER_2020_daily)

    data_.open_data()
    
    #print(information(data_.sites[0])['attributes']['general']['shortName'])


import pandas as pd
import numpy as np


def cell_to_df(cell: np.ndarray, 
               columns: pd.Index | pd.MultiIndex | pd.Series, 
               index: pd.Index | pd.MultiIndex | pd.Series, 
               dtype: str | None):

    """
    for 1d or 2d cell data
    """

    if dtype is None: dtype = cell.dtype

    df                      = pd.DataFrame(cell, columns = columns, index = index, dtype = dtype)
    
    return df


def tile_array_to_list(array, n_times):

    list_df                 = list(array)

    list_tiled              = list_df * n_times

    return list_tiled


def tab_to_array(tabular):

    import pandas as pd

    if isinstance(tabular, pd.Series): 
        
        tabular =  tabular.values
    
    if isinstance(tabular, pd.DataFrame): 
            
        tabular =  tabular.values

        tabular = tabular.flatten()

    return tabular


